# required for datamodel
from django.db import models
from account.models import User, Group
from django.db.models import Q

# required for accessors
from . import stores

from django.template import RequestContext
from django.utils.http import base36_to_int, int_to_base36
from django.utils.crypto import salted_hmac
from django.core.exceptions import ObjectDoesNotExist
from django.core.files import File
from django.core.files.uploadedfile import InMemoryUploadedFile
import urllib.request, urllib.error, urllib.parse
from urllib.parse import urlparse
from datetime import datetime, date
import os
import hashlib
import time
import mimetypes
import tempfile
import re

from medialib import utils
from rest import settings
from rest.models import RestModel, MetaDataBase, MetaDataModel
from rest.decorators import rest_async

from taskqueue.models import Task


MEDIA_FRIENDLY_FILENAMES = settings.get("MEDIA_FRIENDLY_FILENAMES", True)
TASKQUEUE_RENDERING = settings.get("TASKQUEUE_RENDERING", False)

from rest.log import getLogger
logger = getLogger("medialib", filename="medialib.log")

MediaKinds = (
    ('I', 'Image'),
    ('V', 'Video'),
    ('D', 'Audio'),
    ('L', 'Live Video'),
    ('A', 'Live Archive'),
    ('S', 'Video Segment'),
    ('E', 'External Link'),
    ('T', 'Text'),
    ('*', 'Unknown'),
)
RenditionKinds = (
    ('I', 'Image'),
    ('V', 'Video'),
    ('D', 'Audio'),
    ('L', 'Live Stream'),
#   ('A', 'Live Archive *unused*'),
#   ('S', 'Video Segment *unused*'),
    ('N', 'Animated Image'),
    ('E', 'External Link'),
    ('T', 'Text'),
    ('*', 'Unknown'),
)
MediaStates = (
    (0, "Inactive"),
    (1, "Deleted"),
    (50, "Archived"),
    (90, "Hold"),
    (100, "Initial render pending"),
    (101, "Render pending"),
    (102, "Rendering"),
    (110, "Render failed"),
    (111, "Validation failed"),
    (120, "Render ending"),
    (200, "Active"),
)
RenditionParameterKinds = (
    ('S', 'String'),
    ('I', 'Integer'),
    ('F', 'Float'),
    ('B', 'Boolean'),
    ('C', 'Choices'),
)
RenderingEngineStates = (
    ('S', 'Starting'),
    ('I', 'Idle'),
    ('R', 'Rendering'),
    ('0', 'Stopped'),
)

class RenditionParameter(models.Model):
    """
    Available parameters for all renditions
    """
    name = models.CharField(max_length=127, help_text="Parameter name")
    description = models.TextField(null=True, blank=True, help_text="Parameter description")
    required = models.BooleanField(default=False, help_text="Whether is required")
    kind = models.CharField(max_length=1, choices=RenditionParameterKinds, help_text="Type of value")
    choices = models.TextField(null=True, blank=True, help_text="Comma separated list of choices")

    def __str__(self):
        return self.name

    def validate_setting(self, setting):
        if self.kind == 'S':
            return str(setting).strip()
        if self.kind == 'I':
            return int(setting)
        if self.kind == 'F':
            return float(setting)
        if self.kind == 'B':
            if str(setting).lower().strip() in ('1', 'true', 'yes'):
                return True
            elif str(setting).lower().strip() in ('0', 'false', 'no', 'none', ''):
                return False
            raise ValueError("setting is not boolean")
        if self.kind == 'C':
            s = str(setting).lower().strip()
            if s in list(x.lower().strip() for x in self.choices.split(",")):
                return setting.strip()
            raise ValueError("setting is not valid choice")
        raise ValueError("invalid parameter kind")


class RenditionPreset(models.Model):
    """
    Predefined basic rendition settings
    """
    name = models.CharField(max_length=127, help_text="Name of preset")
    short_name = models.SlugField(max_length=32, db_index=True, unique=True, help_text="short name")
    description = models.TextField(null=True, blank=True, help_text="Preset description")
    stage = models.IntegerField(help_text="0-100: run when uploaded, >100: run later")
    parameters = models.ManyToManyField(RenditionParameter, through='RenditionPresetParameterSetting', blank=True, help_text="Default values for parameters")
    configurable_parameters = models.ManyToManyField(RenditionParameter, related_name='applicable_renditions', blank=True, help_text="All parameters that can be set by user")
    default_use = models.CharField(max_length=32, null=True, blank=True, help_text="The default intended use for this rendition types based on this preset")
    applicable_kinds = models.CharField(max_length=16, help_text="Applicable rendition kinds")
    module_name = models.SlugField(max_length=32, db_index=True, help_text="module name")

    def is_distribution(self):
        try:
            return RenditionPresetParameterSetting.objects.get(renditionPreset=self, parameter__name="is_distribution").setting
        except RenditionPresetParameterSetting.DoesNotExist:
            return False

    def __str__(self):
        return self.name

class AccountConfig(models.Model):
    name = models.CharField(max_length=127, help_text="Name config group")
    owner = models.ForeignKey(User, help_text="owner of config", on_delete=models.CASCADE)
    applicable_preset = models.ForeignKey(RenditionPreset, related_name="+", help_text="Applicable preset (info only)", blank=True, null=True, on_delete=models.CASCADE)
    parameters = models.ManyToManyField(RenditionParameter, through='AccountConfigParameterSetting', blank=True, help_text="Values for parameters")

    def __str__(self):
        return self.name

class RenditionDefinition(models.Model):
    """
    User defined rendition types, refined based on RenditionPreset
    """

    class Meta:
        ordering = ["preset__stage", "id"]

    name = models.CharField(max_length=127, db_index=True, help_text="Definition name")
    description = models.TextField(null=True, blank=True, help_text="Definition description")
    preset = models.ForeignKey(RenditionPreset, related_name="+", help_text="Base preset", on_delete=models.CASCADE)
    parameters = models.ManyToManyField(RenditionParameter, through='RenditionDefinitionParameterSetting', help_text="Configured parameters (override values only)")
    active = models.BooleanField(default=True, help_text="Should use for new media")
    use = models.CharField(max_length=32, null=True, blank=True, help_text="The intended use for this rendition type")
    configSet = models.ManyToManyField(AccountConfig, help_text="Additional account config set defaults", blank=True)
    depend = models.CharField(max_length=127, blank=True, db_index=True, help_text="Depends on (use name)")

    def setParameter(self, key, value, kind="I"):
        p = self.renditiondefinitionparametersetting_set.filter(parameter__name=key).last()
        if p is None:
            param = RenditionParameter.objects.filter(name=key).last()
            if param is None:
                param = RenditionParameter(name=key, description=key, kind=kind)
                param.save()
            p = RenditionDefinitionParameterSetting(renditionDefinition=self, parameter=param, setting=value)
            p.save()
        p.setting = value
        p.save()

    def getParameter(self, parameter, item=None):
        if isinstance(parameter, str):
            p = self.renditiondefinitionparametersetting_set.filter(parameter__name=parameter).last()
            if p is not None:
                return p.setting
            p = self.preset.renditionpresetparametersetting_set.filter(parameter__name=parameter).last()
            if p is not None:
                return p.setting
            parameter = RenditionParameter.objects.get(name=parameter)
        if item:
            try:
                return self.mediaitemparametersetting_set.get(parameter=parameter).setting
            except ObjectDoesNotExist:
                pass
        try:
            return self.renditiondefinitionparametersetting_set.get(parameter=parameter).setting
        except ObjectDoesNotExist:
            pass

        for cfg in self.configSet.all().order_by('-id'):
            try:
                return cfg.accountconfigparametersetting_set.get(parameter=parameter).setting
            except ObjectDoesNotExist:
                pass

        try:
            return self.preset.renditionpresetparametersetting_set.get(parameter=parameter).setting
        except ObjectDoesNotExist:
            pass
        return None

    def getParameters(self, item=None):
        params = {}
        for p in self.preset.renditionpresetparametersetting_set.all():
            params[p.parameter.name] = (p.parameter, p.setting, True)
        for cfg in self.configSet.all().order_by('id'):
            for p in cfg.accountconfigparametersetting_set.all():
                params[p.parameter.name] = (p.parameter, p.setting, True)
        for p in self.renditiondefinitionparametersetting_set.all():
            params[p.parameter.name] = (p.parameter, p.setting, False)
        if item:
            for p in item.mediaitemparametersetting_set.all():
                params[p.parameter.name] = (p.parameter, p.setting, False)
        return params

    def __str__(self):
        return self.name

class RenditionSet(models.Model):
    """
    List applicable rendition definitions for a media type
    """

    name = models.CharField(max_length=127, db_index=True, help_text="Set name")
    renditions = models.ManyToManyField(RenditionDefinition, help_text="Renditions used", blank=True)
    kind = models.CharField(max_length=1, choices=MediaKinds, help_text="Kind of media this rendition set is for")
    default_set = models.BooleanField(default=False, help_text="Whether is global default")

    def __str__(self):
        return "%s: %s%s" % (self.kind, self.name, self.default_set and " (default)" or "")

class RenditionPresetParameterSetting(models.Model):
    """
    Key/Value parameter pairs per rendition preset
    """
    class Meta:
        unique_together = (('renditionPreset', 'parameter'),)
    renditionPreset = models.ForeignKey(RenditionPreset, help_text="Preset this pair belongs to", on_delete=models.CASCADE)
    parameter = models.ForeignKey(RenditionParameter, help_text="Parameter", on_delete=models.CASCADE)
    setting = models.CharField(max_length=127, null=True, blank=True, help_text="Value")

    def __str__(self):
        return "%s - %s" % (str(self.renditionPreset), str(self.parameter))

class AccountConfigParameterSetting(models.Model):
    """
    Key/Value parameter pairs per rendition preset
    """
    class Meta:
        unique_together = (('accountConfig', 'parameter'),)
    accountConfig = models.ForeignKey(AccountConfig, help_text="Config this pair belongs to", on_delete=models.CASCADE)
    parameter = models.ForeignKey(RenditionParameter, help_text="Parameter", on_delete=models.CASCADE)
    setting = models.CharField(max_length=127, null=True, blank=True, help_text="Value")

    def __str__(self):
        return "%s - %s" % (str(self.accountConfig), str(self.parameter))

class RenditionDefinitionParameterSetting(models.Model):
    """
    Key/Value parameter pairs per rendition definition
    """
    class Meta:
        unique_together = (('renditionDefinition', 'parameter'),)
    renditionDefinition = models.ForeignKey(RenditionDefinition, help_text="Definition this pair belongs to", on_delete=models.CASCADE)
    parameter = models.ForeignKey(RenditionParameter, help_text="Parameter", on_delete=models.CASCADE)
    setting = models.CharField(max_length=127, null=True, blank=True, help_text="Value")

    def __init__(self, *args, **kwargs):
        try:
            default = kwargs.pop('default')
        except KeyError:
            pass
        else:
            self.default = default
        return super(RenditionDefinitionParameterSetting, self).__init__(*args, **kwargs)

    def __str__(self):
        return "%s - %s" % (str(self.renditionDefinition), str(self.parameter))

class MediaLibrary(models.Model, RestModel):
    """
    Media Library (one per user)
    """
    VIEW_PERMS = ["view_media"]
    SAVE_PERMS = ["manage_media"]
    OWNER_FIELD = "owner"
    name = models.CharField(max_length=127, help_text="name of media library")
    description = models.TextField(null=True, blank=True, help_text="description of media library")
    owner = models.ForeignKey(User, help_text="owner of library", related_name="libraries", blank=True, null=True, default=None, on_delete=models.CASCADE)
    group = models.ForeignKey(Group, related_name="libraries", help_text="group owner of library", blank=True, default=None, null=True, on_delete=models.CASCADE)
    rendition_sets = models.ManyToManyField(RenditionSet, blank=True, help_text="list of RenditionSets used in this library")
    allowed_kinds = models.CharField(max_length=127, null=True, blank=True, help_text="allowed object kinds (null=ALL)")
    acl_default = ((250, None, "R"),)
    acl_override_user = 'owner'

    class Meta:
        permissions = (
            ("can_manage", "Can manage medialib directly"),
        )

    def __str__(self):
        return self.name + ' : ' + str(self.owner)

    def allowed_kinds_list(self):
        if self.allowed_kinds:
            return list(self.allowed_kinds)
        else:
            return list(x[0] for x in MediaKinds)

    def default_store(self):
        try:
            return settings.MEDIALIB_STORE[self.id]
        except (KeyError, AttributeError):
            return settings.MEDIALIB_DEFAULT_STORE


class MediaItem(models.Model, RestModel, MetaDataModel):
    """
    Media Item (a video or image)
    """
    class RestMeta:
        POST_SAVE_FIELDS = ["rendernow"]
        SEARCH_FIELDS = ["name", "description"]
        VIEW_PERMS = ["view_media", "manage_users", "manage_media"]
        SAVE_PERMS = ["manage_media", "manage_users"]
        OWNER_FIELD = "owner"
        GRAPHS = {
            "simple": {
                "fields": [
                    'id',
                    'name',
                    'kind',
                    'state',
                    ('thumbnail_url', "thumbnail"),
                ]
            },
            "basic": {
                "fields": [
                    'id',
                    "created",
                    'name',
                    'description',
                    'kind',
                    'state',
                    ('get_state_display', 'state_display'),
                    'duration',
                    ('thumbnail_url', "thumbnail"),
                    ('smart_renditions', "renditions"),
                ]
            },
            "default": {
                "extra": ["metadata"],
                "graphs": {
                    "self": "basic",
                    "owner": "basic"
                }
            }
        }

    library = models.ForeignKey(MediaLibrary, related_name='items', help_text="Library this item belongs to", null=True, blank=True, default=None, on_delete=models.CASCADE)
    owner = models.ForeignKey(User, help_text="Owner of item", null=True, blank=True, default=None, on_delete=models.CASCADE)
    group = models.ForeignKey(Group, help_text="Owner of item", null=True, blank=True, default=None, on_delete=models.CASCADE)
    name = models.CharField(max_length=127, help_text="Name of item")
    description = models.TextField(blank=True, help_text="Description of item")
    created = models.DateTimeField(auto_now_add=True, editable=False, help_text="When item was created")
    kind = models.CharField(max_length=1, choices=MediaKinds, help_text="Kind of media")
    state = models.IntegerField(choices=MediaStates, help_text="Rendering state of item")
    render_error = models.TextField(null=True, blank=True, help_text="Any render error messages")
    acl_override_user = 'owner'
    acl_parent = 'library'

    def __str__(self):
        return self.name

    def upload__media(self, file, name):
        if not self.name:
            self.name = name
            if file and file.name:
                self.name = file.name
        self.newfile = file

    def set_url(self, value):
        self.newurl = value

    def set_download_url(self, value):
        self.downloadurl = value

    def set_base64(self, value):
        self.base64_data = value

    def __init__(self, *args, **kwargs):
        if 'newfile' in kwargs:
            self.newfile = kwargs.pop('newfile')
        elif 'base64_data' in kwargs:
            self.base64_data = kwargs.pop('base64_data')
            if self.base64_data.startswith("data:"):
                self.base64_data = self.base64_data[self.base64_data.find(',')+1:]
        elif 'newurl' in kwargs:
            self.newurl = kwargs.pop('newurl')
        elif 'downloadurl' in kwargs:
            self.downloadurl = kwargs.pop('downloadurl')
        super(MediaItem, self).__init__(*args, **kwargs)
        self.old_state = self.state

    def onRestCanSave(self, request=None, data=None, extended=None):
        # must throw an exception if cannot save
        return True

    @staticmethod
    def _timestamp():
        return (date.today() - date(2010,1,1)).days

    @staticmethod
    def guessMediaKind(fileobj):
        return utils.guessMediaKind(fileobj)

    def token(self, timestamp=None):
        if timestamp == None:
            timestamp = self._timestamp()
        ts_b36 = int_to_base36(timestamp)
        hash = salted_hmac(settings.SECRET_KEY, ts_b36 + str(self.id) + str(self.created)).hexdigest()[::2]
        return '%s.%s' % (ts_b36, hash)

    def id_token(self, timestamp=None):
        return '%d.%s' % (self.id, self.token(timestamp=timestamp))

    def id_token_perpectual(self):
        return self.id_token(timestamp=0)

    def check_token(self, token):
        if not token:
            return False
        ts_b36 = token.split(".")[0]
        timestamp = base36_to_int(ts_b36)
        nowts = self._timestamp()
        if timestamp > 0 and (timestamp > nowts or timestamp < nowts-2):
            return False
        return "%d.%s" % (self.id, token) == self.id_token(timestamp)

    def default_store(self):
        if self.library:
            return self.library.default_store()
        return settings.MEDIALIB_DEFAULT_STORE

    def s3_store(self):
        return f's3://{settings.AWS_S3_BUCKET}'

    def set_rendernow(self, value):
        if self.kind == "*":
            self.saveState(200)
            return
        self.new_render(force_task=(self.kind == "V"))

    def uses(self):
        """
        List uses available for this item
        """
        return list(self.renditions.values('use').distinct().values_list('use', flat=True))

    def rerender(self):
        self.deleteRenditions()
        self.saveState(100)
        # {'original': newfile}
        return self.new_render()

    def async_render(self, files={}):
        return self.new_render()

    def new_render(self, files={}, force_task=False):
        if TASKQUEUE_RENDERING or force_task:
            Task.Publish("medialib", "on_render", {"id": self.id, "name": self.name, "kind": self.kind}, channel="tq_app_handler_medialib")
        else:
            self.createRenditions()
        return True

    def createRenditions(self, files={}):
        self.saveState(101) #  rendering
        logger.info("create renditions, {}:{}".format(self.name, self.kind))
        render = __import__('medialib.render', globals(), locals(), ['render'])
        res = render.render(self, 1, files=files, log=logger.info)
        if not res:
            self.saveState(110)
            self._renderFailed(files)
        else:
            self.saveState(200)

    def _renderFailed(self, files):
        err = self.render_error
        if files:
            faildir = os.environ.get('TMPDIR', '/tmp') + "/validate_fail"
            try:
                os.mkdir(faildir)
            except OSError:
                pass
            for f in files:
                filename = faildir + "/" + str(int(time.time())) + "_" + f
                try:
                    outf = open(filename + ".dat", 'wb')
                    fp = files[f]
                    if hasattr(fp, "chunks"):
                        for chunk in fp.chunks():
                            outf.write(chunk)
                    else:
                        chunk = fp.read(1024)
                        while chunk:
                            outf.write(chunk)
                            chunk = fp.read(1024)
                    outf.close()
                    outf = open(filename + ".txt", 'w')
                    import pprint
                    outf.write(pprint.pformat(vars(files[f])))
                    outf.write(pprint.pformat(vars(self)))
                    outf.close()
                except OSError:
                    pass
        if self.state == 111:
            # ignore errors
            try:
                self.deleteRenditions()
            except Exception as err:
                logger.exception(err)
        # raise Exception(err)
        # Generate Event

    def guessKind(self):
        filename = self.name
        if getattr(self, "newfile", None):
            filename = self.newfile.name
        if getattr(self, "base64_data", None):
            # base64 lets attempt to read the first few bytes
            kind = utils.guessBase64Kind(self.base64_data)
            if kind != None:
                self.ext = kind
                if kind in ["png", "jpg", "jpeg", "jfif", "bmp", "gif", "tif"]:
                    self.kind = "I"
                    return
        (mt, _) = mimetypes.guess_type(filename)
        if mt == None:
            self.kind = '*'
        elif mt[:6] == 'image/':
            self.kind = 'I'
        elif mt[:6] == 'video/':
            self.kind = 'V'
        elif mt == 'application/x-shockwave-flash':
            self.kind = 'V'
        else:
            self.kind = '*'

    def saveState(self, state):
        if self.state != state:
            self.state = state
            return super(MediaItem, self).save()
        return True

    def save(self, *args, **kwargs):
        if self.pk is None:
            return self.saveNew(*args, **kwargs)
        return super(MediaItem, self).save(*args, **kwargs)

    def saveNew(self, *args, **kwargs):
        """
        When saving MediaItem set newfile attribute to upload new file and start rendering
        """
        # set state to render-pending if new upload
        has_fields = hasattr(self, 'newfile') or hasattr(self, 'newurl') or hasattr(self, 'downloadurl') or hasattr(self, 'base64_data')
        state = self.state
        if self.state is None or self.state >= 100 or has_fields:
            self.state = 100

        if not self.kind:
            self.guessKind()

        if self.kind and self.library and self.library.allowed_kinds and not self.kind in self.library.allowed_kinds:
            raise Exception("Invalid media type for this library")

        if not self.library:
            lib, created = MediaLibrary.objects.get_or_create(name="default", description="default", owner=User.objects.first())
            self.library = lib

        ret = super(MediaItem, self).save(*args, **kwargs)

        rkind = self.kind
        if rkind in ('A','S',):
            rkind = 'V'
        if hasattr(self, 'newfile'):
            newfile = self.newfile
            delattr(self, 'newfile')
            rendition = MediaItemRendition(mediaitem=self, name="Original", use='original', kind=rkind, is_original=True)
            rendition.upload(newfile)
            rendition.save()
            if self.state >= 100:
                if rkind == 'V':
                    self.async_render({'original': newfile})
                else:
                    self.new_render({'original': newfile})
        elif hasattr(self, 'base64_data'):
            fileName, ext = os.path.splitext(self.name)
            if hasattr(self, "ext"):
                ext = self.ext
            # print "media suffix: " + ext
            # print "media kind: " + self.kind
            newfile = tempfile.NamedTemporaryFile(suffix=ext)
            newfile.write(utils.fromBase64ToBytes(self.base64_data))
            newfile.seek(0)
            delattr(self, 'base64_data')
            rendition = MediaItemRendition(mediaitem=self, name="Original", use='original', kind="I", is_original=True)
            rendition.upload(newfile)
            rendition.save()
            if self.state >= 100:
                self.new_render({'original': newfile})
        elif hasattr(self, 'downloadurl'):
            # this needs to change to a background task via a stage of rendering
            # original should be the URL
            newurl = self.downloadurl
            delattr(self, 'downloadurl')
            res = urllib.request.urlopen(newurl)
            fileName, ext = os.path.splitext(newurl)
            newfile = tempfile.NamedTemporaryFile(suffix=ext)
            # newfile.name = self.name
            newfile.write(res.read())
            newfile.seek(0)
            rendition = MediaItemRendition(mediaitem=self, name="Original", use='original', kind=rkind, is_original=True)
            rendition.upload(newfile)
            rendition.save()
            rendition = MediaItemRendition(mediaitem=self, name="OriginalSource", use='source', url=newurl, bytes=0, kind=rkind)
            rendition.save()
            if self.state >= 100:
                self.new_render({'original': newfile})
        elif hasattr(self, 'newurl'):
            newurl = self.newurl
            delattr(self, 'newurl')
            rendition = MediaItemRendition(mediaitem=self, name="Original", use='original', url=newurl, bytes=0, kind=rkind, is_original=True)
            rendition.save()
            if hasattr(self, 'newurl_local'):
                meta = MediaMeta(rendition=rendition, key="url_local", value=self.newurl_local)
                meta.save()
            if self.state >= 100:
                self.new_render()
        return ret

    def video_rendition_list(self):
        return self.renditions.filter(Q(use='video')|Q(use="original")).order_by('width')

    def original(self):
        """
        Get original rendition
        """
        try:
            if self.kind == 'A':
                return self.renditions.exclude(kind='L').filter(is_original=True).order_by('created')[0]
            return self.renditions.filter(is_original=True).order_by('-created')[0]
        except IndexError:
            return None

    def still(self):
        """
        Get still rendition
        """
        if self.kind == 'I':
            return self.original()
        try:
            return self.renditions.filter(use='still').order_by('-created')[0]
        except IndexError:
            return MediaItemRendition(name="Preview not yet available", url="/static/img/noimage.gif", fake=True)

    def poster_url(self, request=None):
        return self.get('poster', 'still', 'original', 'noimage').view_url(request=request, expires=None)

    def get(self, *args):
        """
        Get first matching rendition
        """
        for use in args:
            if use == 'noimage':
                return MediaItemRendition(name="Rendition not yet available", url="/static/img/noimage.gif", fake=True)
            try:
                return self.renditions.filter(use=use).order_by('-created')[0]
            except IndexError:
                pass
        return None

    def thumbnail_animated(self):
        """
        Get animated thumbnail rendition
        """
        return self.renditions.filter(use='thumbnail-animated').order_by('-created').first()

    def isImage(self):
        org = self.original()
        if org is not None:
            return org.kind == "I"
        return False

    def isAnimated(self):
        org = self.original()
        if org is not None:
            return org.kind == "N"
        return False

    def animated_url(self):
        if self.kind == "V":
            thumb = self.thumbnail_animated()
            if thumb:
                return thumb.view_url_nonexpire()
        elif self.kind == "I":
            org = self.original()
            if org.kind == "N":
                return org.view_url_nonexpire()
        return None

    def image_small(self):
        """
        Get small tumbnail rendition
        """
        try:
            return self.renditions.filter(use='image', width__lt=200).order_by('-created')[0]
        except IndexError:
            pass
        try:
            return self.renditions.filter(use='image').order_by('-created')[0]
        except IndexError:
            pass
        return self.still()

    def thumbnail_url(self, width=320):
        thumb = self.renditions.filter(kind="I", use="thumbnail", width__lte=width).order_by('-width').first()
        if thumb:
            return thumb.view_url_nonexpire()
        return None

    def thumbnail_small(self):
        """
        Get small tumbnail rendition
        """
        try:
            return self.renditions.filter(use='thumbnail', width__lt=80).order_by('-created')[0]
        except IndexError:
            pass
        try:
            return self.renditions.filter(use='thumbnail').order_by('-created')[0]
        except IndexError:
            pass
        if self.kind == '*':
            return MediaItemRendition(name="Unknown Attachment", url="/static/img/unknown.gif", fake=True)
        return self.still()

    def thumbnail_large(self):
        """
        Get large thumbnail rendition
        """
        try:
            return self.renditions.filter(use='thumbnail', width__gte=80).order_by('-created')[0]
        except IndexError:
            pass
        try:
            return self.renditions.filter(use='thumbnail').order_by('-created')[0]
        except IndexError:
            pass
        if self.kind == '*':
            return MediaItemRendition(name="Unknown Attachment", url="/static/img/unknown.gif", fake=True)
        return self.still()

    def image_larger(self, size):
        """
        Get any image with width same or larger than <size>
        """
        try:
            return self.renditions.filter(use__in = ('thumbnail', 'image'), width__gte=size).order_by('width', '-created')[0]
        except IndexError:
            pass

        return self.still()

    def video_url(self, quality="480p", request=None):
        video = self.renditions.filter(kind="V", name=quality).order_by('-id').first()
        if video is None:
            video = self.original()
            if video is None:
                return None
        return video.view_url(request=request, expires=None)

    def image_url(self, request=None):
        image = self.renditions.filter(kind="I", use="thumbnail").order_by('-width').first()
        if image is None:
            image = self.original()
            if image is None:
                return None
        url = image.view_url(request=request, expires=None)
        if url and not url.startswith("http"):
            if url.startswith("/"):
                url = url[1:]
            return "{}{}".format(settings.BASE_URL_SECURE, url)
        return url

    def original_url(self, request=None):
        org = self.original()
        if org is not None:
            return org.view_url(request=request, expires=None)
        return self.url(request=request)

    def url(self, size=None, request=None):
        if self.kind == "V":
            return self.video_url(request=request)
        elif self.kind == "I":
            return self.image_url(request=request)
        org = self.original()
        if org is not None:
            return org.view_url(request=request, expires=None)
        return None

    def getImageRendition(self, width=640, name=None, use="thumbnail", request=None, flat=True):
        image = None
        if name is None:
            image = self.renditions.filter(kind="I", use=use, width__lte=width).order_by('-width').first()
        else:
            image = self.renditions.filter(kind="I", name=name).order_by('-width').first()
        if image is None:
            return None
        if not flat:
            return image
        return {
            "kind": image.kind,
            "content_type": mimetypes.guess_type(image.url)[0],
            "bytes":image.bytes,
            "url":image.view_url(request=request, expires=None, is_secure=True),
            "width":image.width,
            "height":image.height
        }

    def getVideoRendition(self, height=640, request=None, flat=True):
        video = self.renditions.filter(kind="V", use="video", height__lte=height).order_by('-height').first()
        if not flat:
            return video
        if video:
            return {
                "bytes":video.bytes,
                "url":video.view_url(request=request, expires=None, is_secure=True),
                "width":video.width,
                "height":video.height,
                "use":"video"
            }
        return None

    def getOriginalRendition(self, request=None):
        orig = self.original()
        if orig is None:
            return None
        return {
            "kind": orig.kind,
            "content_type": mimetypes.guess_type(self.name)[0],
            "bytes":orig.bytes,
            "url":orig.view_url(request=request, expires=None, is_secure=True),
            "width":orig.width,
            "height":orig.height
        }

    def image_renditions(self, request=None):
        return {
            "square":self.getImageRendition(name="square thumbnail", request=request),
            "medium":self.getImageRendition(width=620, request=request),
            "large":self.getImageRendition(width=1020, request=request),
            "original":self.getOriginalRendition(request=request)
        }

    def remote_renditions(self, request=None):
        return {
            "websnap":self.getImageRendition(use="websnap", width=2000, request=request),
            "medium":self.getImageRendition(width=2000, request=request),
            "large":self.getImageRendition(width=2000, request=request),
            "original":self.getOriginalRendition(request=request)
        }

    def video_renditions(self, request=None):
        animated = self.thumbnail_animated()
        res = {
            "low":self.getVideoRendition(height=480, request=request),
            "medium":self.getVideoRendition(height=720, request=request),
            "high":self.getVideoRendition(height=1080, request=request),
            "original":self.getOriginalRendition(request=request)
        }
        if animated:
            res["animated"] = {
                "bytes":animated.bytes,
                "url":animated.view_url(request=request, expires=None, is_secure=True),
                "width":animated.width,
                "height":animated.height,
                "use":"gif"
            }
        poster = self.get('poster', 'still', 'original', 'noimage')
        if poster:
            res["poster"] = {
                "bytes":poster.bytes,
                "url":poster.view_url(request=request, expires=None, is_secure=True),
                "width":poster.width,
                "height":poster.height,
                "use":"poster"
            }
        thumb = self.getImageRendition(name="square thumbnail", request=request)
        if thumb:
            res["square_thumbnail"] = thumb

        thumb = self.getImageRendition(name="small thumbnail", request=request)
        if thumb:
            res["small_thumbnail"] = thumb

        thumb = self.getImageRendition(name="medium thumbnail", request=request)
        if thumb:
            res["large_thumbnail"] = thumb

        thumb = self.getImageRendition(name="video still", request=request)
        if thumb:
            res["video_still"] = thumb
        return res

    def smart_renditions(self, request=None):
        if self.kind == "I":
            return self.image_renditions(request=request)
        elif self.kind == "V":
            return self.video_renditions(request=request)
        elif self.kind == "E":
            return self.remote_renditions(request=request)
        return { "original": self.getOriginalRendition(request=request) }

    def duration(self):
        org = self.original()
        if org is not None:
            return org.get_meta('duration', 0.0, float)
        return 0.0

    def bytes(self):
        org = self.original()
        if org is not None:
            return org.bytes
        return 0

    def live_state_ingesting(self):
        return self.kind == 'L' and self.state > 100 and self.state < 110
    def live_state_available(self):
        return self.live_state_ingesting() or (self.state == 200)
    def live_ingest_url(self):
        try:
            return "/".join(self.original().url.split("/")[:-1])
        except AttributeError:
            return None
    def live_local_url(self):
        orig = self.original()
        if not orig:
            return None
        try:
            url = orig.mediameta_set.get(key='url_local').value
        except MediaMeta.DoesNotExist:
            url = orig.url

        try:
            return "/".join(url.split("/")[:-1])
        except AttributeError:
            return None
    def live_stream_id(self):
        try:
            return self.original().url.split("/")[-1]
        except AttributeError:
            return None

    def addRendtion(self, name, use, kind, url, width=0, height=0):
        #render = __import__('medialib.render', globals(), locals(), ['render'])
        rendition = MediaItemRendition(mediaitem=self, name=name,
            use=use, url=url,
            bytes=0, width=width, height=height,
            kind=kind, is_original=False)
        rendition.save()
        return rendition

    def deleteRenditions(self, keep_original=True):
        """
        Delete rendition (including stored file)
        """
        for r in self.renditions.all():
            if keep_original and r.use == "original":
                continue
            r.delete()

    def updateFromURL(self, url):
        url = url.strip()
        try:
            u = urllib.request.urlopen(url)
            if not u.code == 200:
                return None
        except IOError:
            return None

        self.deleteRenditions()
        self.state = 100
        res = urllib.request.urlopen(url)
        fileName, ext = os.path.splitext(url)
        newfile = tempfile.NamedTemporaryFile(suffix=ext)
        # newfile.name = self.name
        newfile.write(res.read())
        newfile.seek(0)
        rendition = MediaItemRendition(mediaitem=self, name="Original", use='original', kind=self.kind, is_original=True)
        rendition.upload(newfile)
        rendition.save()

        rendition = MediaItemRendition(mediaitem=self, name="OriginalSource", use='source', url=url, bytes=0, kind=self.kind, is_original=True)
        rendition.save()

        self.save()
        self.new_render({'original': newfile})

    @staticmethod
    def CreateFromURL(url, owner, library=None, kind=None):
        try:
            u = urllib.request.urlopen(url)
            if not u.code == 200:
                return None
        except IOError:
            return None

        filename = utils.getFileNameFromURL(url)
        if not kind:
            kind = utils.guessMediaKindByName(filename)

        # our first library should be our private library
        if not library:
            library = MediaLibrary.objects.filter(owner=owner).first()
            if not library:
                library = MediaLibrary(name='Default Library', owner=owner)
                library.save()

        media = MediaItem(library=library, name=filename, owner=owner, kind=kind, downloadurl=url)
        media.save()
        return media

    @staticmethod
    def on_upload_s3(request):
        filename = request.DATA.get("filename")
        filesize = request.DATA.get("filesize")
        filetype = request.DATA.get("filetype")
        kind = utils.guessMediaKindByName(filename)
        if kind is None:
            kind = "*"
        obj = MediaItem(name=filename, owner=request.member, kind=kind, group=request.group)
        obj.save()
        rendition = MediaItemRendition(
            mediaitem=obj, name="Original", bytes=filesize,
            use='original', kind=kind, is_original=True)
        rendition.save()
        rendition.url = rendition.generateURL(filename, store=obj.s3_store())
        rendition.save()
        return dict(url=rendition.generateUploadURL(filetype), id=obj.pk)


class MediaItemMetaData(MetaDataBase):
    parent = models.ForeignKey(MediaItem, related_name="properties", on_delete=models.CASCADE)

class MediaItemRef(models.Model, RestModel):
    class RestMeta:
        GRAPHS = {
            "default": {
                "graphs":{
                    "media":"basic",
                    "generic__component":"basic",
                }
            },
            "list": {
                "graphs":{
                    "media":"basic",
                }
            },
            "simple": {
                "graphs":{
                    "media":"basic"
                }
            },
        }
    created = models.DateTimeField(auto_now_add=True, editable=False)
    media = models.ForeignKey(MediaItem, related_name="references", on_delete=models.CASCADE)
    component = models.CharField(max_length=200, blank=True, null=True, default=None, db_index=True)
    component_id = models.IntegerField(db_index=True)

class CuePoint(models.Model):
    """
    Cue points within media item
    """
    item = models.ForeignKey(MediaItem, on_delete=models.CASCADE)
    start = models.FloatField()
    end = models.FloatField(null=True, blank=True)
    subitem = models.ForeignKey(MediaItem, related_name="cueparent", on_delete=models.CASCADE)


class CuePointMeta(models.Model):
    """
    MetaData on MediaItemRendition
    """

    cue = models.ForeignKey(CuePoint, on_delete=models.CASCADE)
    key = models.CharField(max_length=127, db_index=True, help_text="Key")
    value = models.TextField(null=True, blank=True, help_text="Value")

    def __str__(self):
        return self.key


class MediaItemRendition(models.Model):
    """
    A rendition of a media item
    """

    mediaitem = models.ForeignKey(MediaItem, related_name='renditions', help_text="Item this rendition is for", on_delete=models.CASCADE)
    name = models.CharField(max_length=127, help_text="Name of rendition (copied from definition)")
    use = models.CharField(max_length=32, null=True, blank=True, help_text="The intended use for this media item (copied from definition)")
    rendition_definition = models.ForeignKey(RenditionDefinition, null=True, help_text="The rendition type that generated this rendition", on_delete=models.CASCADE)
    url = models.CharField(max_length=255, help_text="internal url for rendition")
    created = models.DateTimeField(auto_now_add=True, editable=False, help_text="When item was created")
    width = models.IntegerField(null=True, help_text="Width (if available)")
    height = models.IntegerField(null=True, help_text="Height (if available)")
    bytes = models.BigIntegerField(help_text="Size of rendition file")
    kind = models.CharField(max_length=1, choices=RenditionKinds, help_text="Kind of rendition")
    is_original = models.BooleanField(default=False, help_text="Whether is original of kind")

    @property
    def key(self):
        u = urlparse(self.url)
        return u.path.lstrip('/')

    def __init__(self, *args, **kwargs):
        self._fake = kwargs.pop('fake', False)
        super(MediaItemRendition, self).__init__(*args, **kwargs)

    def __bool__(self):
        return not self._fake

    def __str__(self):
        try:
            return str(self.mediaitem) + ": " + self.name
        except MediaItem.DoesNotExist:
            return self.name

    def ext(self):
        """
        File extension of rendition
        """
        try:
            ext = self.url.split('.')[-1]
            if '/' in ext: return ''
            return ext
        except IndexError:
            return ''

    def generateUploadURL(self, filetype):
        from medialib.stores import s3
        return s3.generate_upload_url(self.url, filetype)

    def generateFriendlyURL(self, name, prefix="", store=None):
        from rest import helpers as rh
        rh.log_error("generateFriendlyURL", name, self.use, self.mediaitem.name, " ")
        if store is None:
            store = self.mediaitem.default_store()
        if self.use != "original":
            if "." in self.mediaitem.name:
                orig_name = self.mediaitem.name.split('.')
                orig_ext = orig_name.pop()
                orig_name = "".join(orig_name)
                if "." in name:
                    ext = name.split('.')[-1]
                else:
                    ext = orig_ext
                name = f"{orig_name}__{self.name}.{ext}"
        rh.log_error("--- now ---", name, "  ")
        name = name.replace(" ", "_")
        name = re.sub(r'[^A-Za-z0-9._-]', '', name)
        paths = []
        paths.append(store)
        paths.append("/")
        paths.append(int_to_base36(self.mediaitem.pk))
        paths.append("/")
        paths.append(name)
        return "".join(paths)

    def generateURL(self, name, prefix="", store=None):
        if MEDIA_FRIENDLY_FILENAMES:
            return self.generateFriendlyURL(name, prefix, store)

        paths = []
        paths.append(store)
        paths.append("/{}/".format(int_to_base36(self.mediaitem.pk)))
        path = "".join(paths)
        if self.rendition_definition is not None:
            paths.append(int_to_base36(self.rendition_definition.pk))
        else:
            paths.append("0")
        paths.append("_")
        paths.append(utils.toMD5(path, int(time.time())))
        paths.append(".")
        ext = "dat"
        if "." in name:
            ext = name.split('.')[-1]
        elif hasattr(self.mediaitem, "ext"):
            ext = self.mediaitem.ext
        paths.append(ext)
        # print(paths)
        return "".join(paths)

    def upload(self, fp, prefix=""):
        """
        Upload rendition to non-volatile storage
        """
        doclose = False
        if type(fp) == str:
            doclose = True
            fp = open(fp)

        self.url = self.generateURL(fp.name, prefix)
        stores.upload(self.url, fp, background=True)
        try:
            self.bytes = fp.size
        except AttributeError:
            self.bytes = os.fstat(fp.fileno()).st_size

        if doclose:
            fp.close()
        else:
            try:
                if self.fp:
                    return
            except  AttributeError:
                pass
            self.fp = fp
            try:
                self.filename = fp.name
            except AttributeError:
                pass

    def view_url(self, expires=3600, request=None, is_secure=None):
        """
        Get the URL of rendition that can be used for viewing
        """
        if is_secure is None:
            is_secure = request and hasattr(request, 'is_secure') and request.is_secure()

        try:
            if self.rendition_definition.preset.short_name == "hls":
                url = "/medialib/hls/%d.%s/index.m3u8" % (self.pk, self.mediaitem.token())
                if not self.get_meta('final'):
                    url += "?live=1"
                return url
            elif self.use == "youtube":
                pre = "http"
                if is_secure:
                    pre = "https"
                return "{0}://youtu.be/{1}".format(pre, self.url)
        except ObjectDoesNotExist:
            pass
        except AttributeError:
            pass

        return stores.view_url(self.url, expires=expires, is_secure=is_secure, request=request)

    def view_url_nonexpire(self, request=None, is_secure=True):
        return self.view_url(expires=None, request=request, is_secure=is_secure);

    def rtmp_url(self, expires=3600, request=None):
        """
        Get the URL of rendition that can be used for viewing
        """
        try:
            if self.rendition_definition.preset.short_name == "hls":
                return None
        except ObjectDoesNotExist:
            pass
        except AttributeError:
            pass
        if request and hasattr(request, 'is_secure') and request.is_secure():
            is_secure = True
        else:
            is_secure = False
        return stores.rtmp_url(self.url, expires=expires, is_secure=is_secure, request=request)

    def get_local_url(self):
        """
        Get local url for rendition
        """
        try:
            if self.rendition_definition.preset.short_name == "hls":
                return "http://" + settings.HOSTNAME + "/medialib/hls/%d.%s/index.m3u8" % (self.pk, self.mediaitem.token())
        except ObjectDoesNotExist:
            pass
        except AttributeError:
            pass
        extra = ""
        if stores.type(self.url) == "rtmpstore":
            if self.mediaitem.kind == "L":
                extra = " live=1 buffer=200"
            else:
                extra = " buffer=1"
        try:
            return stores.view_url(self.mediameta_set.get(key='url_local').value) + extra
        except MediaMeta.DoesNotExist:
            return stores.view_url(self.url) + extra

    def check_file(self):
        """
        Check if file exists in backend
        """
        if self.kind == "E":
            return True

        if self.url[-1] == "/":
            raise Exception("check_file on directory")
        return stores.exists(self.url) != False

    def read_all(self):
        """
        read the rendition
        """
        return stores.get_file(self.url, fp)

    def get_file(self, fp=None):
        """
        Fetch the rendition into file
        """
        if self.url[-1] == "/":
            raise Exception("getfile on directory")
        if fp is None:
            fp = tempfile.NamedTemporaryFile(suffix=('.' + self.ext()), prefix="{0}_".format(self.id))
        stores.get_file(self.url, fp)
        fp.seek(0)
        return fp

    def delete(self):
        """
        Delete rendition (including stored file)
        """
        if self.url.startswith(self.mediaitem.default_store()):
            stores.delete(self.url)
        return super(MediaItemRendition, self).delete()

    def set_meta(self, key, value):
        (meta, _) = MediaMeta.objects.get_or_create(rendition=self, key=key)
        meta.value = value
        meta.save()
        return meta

    def get_meta(self, key, dflt=None, type=None):
        try:
            ret = self.mediameta_set.get(key=key).value
        except MediaMeta.DoesNotExist:
            ret = dflt
        if type:
            try:
                ret = type(ret)
            except ValueError:
                ret = dflt
        return ret

    def bitrate(self):
        rate = self.rendition_definition.getParameter('video_bitrate')
        if rate:
            return int(rate) * 1024
        if self.bytes:
            dur = self.get_meta('duration')
            if dur:
                return bytes / dur
        return None

    def updateFromURL(self, url):
        url = url.strip()
        try:
            u = urllib.request.urlopen(url)
            if not u.code == 200:
                return None
        except IOError:
            return None

        res = urllib.request.urlopen(url)
        fileName, ext = os.path.splitext(url)
        newfile = tempfile.NamedTemporaryFile(suffix=ext)
        # newfile.name = self.name
        newfile.write(res.read())
        newfile.seek(0)
        self.upload(newfile)
        self.save()
        # self.new_render({'original': newfile})


class MediaMeta(models.Model):
    """
    MetaData on MediaItemRendition
    """

    rendition = models.ForeignKey(MediaItemRendition, help_text="Rendition this metadata belongs to", on_delete=models.CASCADE)
    key = models.CharField(max_length=127, db_index=True, help_text="Key")
    value = models.TextField(null=True, blank=True, help_text="Value")

    def __str__(self):
        return self.key

def _next_priority():
    prios = RenderInstance.objects.all().order_by('-priority').values_list('priority', flat=True)[:1]
    if len(prios) == 0:
        return 100
    elif not prios[0]:
        return 100
    else:
        return prios[0] + 1

class RenderInstance(models.Model):
    """
    Rendering engine instances
    """

    instance_id = models.CharField(max_length=16, db_index=True, help_text="Rendering engine ID")
    state = models.CharField(max_length=1, db_index=True, choices=RenderingEngineStates, help_text="Current state of engine")
    started = models.DateTimeField(auto_now_add=True, editable=False, help_text="When rendering was started")
    last_checkin = models.DateTimeField(default=datetime.now, help_text="time instance last checked in")
    shutdown = models.DateTimeField(null=True, blank=True, help_text="Time of instance shutdown")
    priority = models.IntegerField(default=_next_priority, null=True, blank=True, help_text="Render instance priority")
    rendering = models.ForeignKey(MediaItem, null=True, blank=True, related_name='render_instance', help_text="Media item currently being rendered", on_delete=models.CASCADE)
    message = models.TextField(null=True, blank=True, help_text="Render message")

    def __str__(self):
        return self.instance_id

    def last_message(self):
        try:
            return self.message.strip().split('\n')[-1]
        except (IndexError, AttributeError):
            return ''

class RenditionSegment(models.Model):
    """
    Segment of media rendition
    """
    rendition = models.ForeignKey(MediaItemRendition, related_name='segments', help_text="Rendition this metadata belongs to", on_delete=models.CASCADE)
    segment = models.IntegerField(help_text="Segment number")
    start = models.FloatField(db_index=True, help_text="Start time offset")
    end = models.FloatField(db_index=True, help_text="End time offset")
    duration = models.FloatField(help_text="Segment duration")
    bytes = models.IntegerField(help_text="Segment size")
    url = models.CharField(max_length=255, help_text="internal url for rendition")

    def __str__(self):
        return "%s %d" % (self.rendition, self.segment)

class MediaItemParameterSetting(models.Model):
    """
    Key/Value parameter pairs per media item
    """
    class Meta:
        unique_together = (('item', 'parameter'),)
    item = models.ForeignKey(MediaItem, help_text="Media Item this pair belongs to", on_delete=models.CASCADE)
    parameter = models.ForeignKey(RenditionParameter, help_text="Parameter", on_delete=models.CASCADE)
    setting = models.CharField(max_length=127, null=True, blank=True, help_text="Value")

    def __str__(self):
        return "%s - %s" % (str(self.item), str(self.parameter))





from django.db import dm as dm
from rest import dm as rm

import mistune
from .renderers import WikiRenderer


class FaqItem(dm.Model, rm.RestModel):
    class Meta:
        ordering = ['-modified']

    class RestMeta:
        SEARCH_FIELDS = ["subject"]

    created = dm.DateTimeField(auto_now_add=True, editable=False)
    modified = dm.DateTimeField(auto_now=True, db_index=True)

    owner = dm.ForeignKey(
        "account.Member", related_name="+",
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    group = dm.ForeignKey(
        "account.Group", related_name="+", 
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    kind = dm.CharField(max_length=124, db_index=True)
    subject = dm.TextField(db_index=True)
    description = dm.TextField(blank=True, null=True, default=None)
    html = dm.TextField(blank=True, null=True, default=None)

    def set_description(self, value):
        self.description = value
        self.html = self.toHTML()

    def toHTML(self):
        renderer = WikiRenderer()
        md_engine = mistune.Markdown(renderer=renderer)
        html = md_engine.render(self.description)
        return html


class FaqMedia(dm.Model, rm.RestModel):
    created = dm.DateTimeField(auto_now_add=True, editable=False)
    modified = dm.DateTimeField(auto_now=True, db_index=True)

    faq = dm.ForeignKey(FaqItem, blank=True, null=True, default=None, related_name="media", on_delete=dm.CASCADE)

    owner = dm.ForeignKey(
        "account.Member", related_name="+",
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    media = dm.ForeignKey("medialib.MediaItem", related_name="+", null=True, default=None, on_delete=dm.CASCADE)



# Generated by Django 4.1.4 on 2023-11-20 03:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incident', '0009_incident_reporter_ip'),
    ]

    operations = [
        migrations.AddField(
            model_name='incident',
            name='category',
            field=models.CharField(db_index=True, default=None, max_length=124, null=True),
        ),
        migrations.AddField(
            model_name='incident',
            name='component_id',
            field=models.IntegerField(blank=True, db_index=True, default=None, null=True),
        ),
    ]



from objict import objict
from datetime import datetime
import re
from .. import models as am
from location.models import GeoIP

IGNORE_RULES = [
    "100020"
]

LEVEL_REMAP_BY_RULE = {
    5402: 7,
    5710: 5
}

NGINX_PARSE_PATTERN = re.compile(
    r'(?P<src_ip>\d+\.\d+\.\d+\.\d+) - - \[(?P<http_time>.+?)\] '
    r'(?P<http_method>\w+) (?P<http_url>.+?) (?P<http_protocol>[\w/.]+) '
    r'(?P<http_status>\d+) (?P<http_bytes>\d+) (?P<http_referer>.+?) '
    r'(?P<user_agent>.+?) (?P<http_elapsed>\d\.\d{3})'
)

def parse_nginx_line(line):
    if "\n" in line:
        for l in line.split('\n'):
            match = NGINX_PARSE_PATTERN.match(l)
            if match:
                return match.groupdict()
        return None
    match = NGINX_PARSE_PATTERN.match(line)
    if match:
        return match.groupdict()
    return None


def removeNonAscii(input_str, replacement=''):
    """
    Replace all non-ASCII characters and escaped byte sequences in the input string with a specified string.
    
    Args:
    input_str (str): The string to process.
    replacement (str): The string to use as a replacement for non-ASCII characters and escaped byte sequences.

    Returns:
    str: The processed string with non-ASCII characters and byte sequences replaced.
    """
    # Replace escaped byte sequences with the replacement string
    cleaned_str = re.sub(r'\\x[0-9a-fA-F]{2}', replacement, input_str)
    # Replace non-ASCII characters with the replacement string
    return ''.join(char if (32 <= ord(char) < 128 or char in '\n\r\t') else f"<r{str(ord(char))}>" for char in cleaned_str)


def extractURL(text):
    match = re.search(r"(GET|POST|DELETE|PUT)\s+(https?://[^\s]+)\s+HTTP/\d\.\d", text)
    if match:
        return match.group(2)
    return None


def extractDomain(text):
    match = re.search(r"https?://([^/:]+)", text)
    if match:
        return match.group(1)
    return None


def extractIP(text):
    match = re.search(r"\b(?:[0-9]{1,3}\.){3}[0-9]{1,3}\b", text)
    if match:
        return match.group(1)
    return None


def extractUrlPath(text):
    match = re.search(r"https?://[^/]+(/[^?]*)", text)
    if match:
        return match.group(1)
    return None


def extractUserAgent(text):
    # this only works if the referrer is '-'
    match = re.search(r"' - (.+?) \d+\.\d+ \d+'", text)
    if match:
        return match.group(1)
    return None


def extractMetaData(alert):
    return parse_alert_metadata(alert)


DEFAULT_META_PATTERNS = {
    "src_ip": re.compile(r"Src IP: (\S+)"),
    "src_port": re.compile(r"Src Port: (\S+)"),
    "user": re.compile(r"User: (\S+)"),
    "path": re.compile(r"request: (\S+ \S+)"),
    "http_server": re.compile(r"server: (\S+),"),
    "http_host": re.compile(r"host: (\S+)"),
    "http_referrer": re.compile(r"referrer: (\S+),"),
    "client": re.compile(r"client: (\S+),"),
    "upstream": re.compile(r"upstream: (\S+),")
}


def parse_alert_metadata(alert):
    patterns = DEFAULT_META_PATTERNS
    if alert.rule_id == "2501":
        match = re.search(r'user (\w+) (\d+\.\d+\.\d+\.\d+) port (\d+)', alert.text)
        if match:
            return dict(username=match.group(1), src_ip=match.group(2), src_port=match.group(3))
    elif alert.rule_id.startswith("311") or alert.rule_id in ["31516", "31508", "31516"]:
        data = parse_nginx_line(alert.text)
        if data:
            return data
    elif alert.rule_id in ["31301"]:
        match = re.search(r'\((?P<error_code>\d+): (?P<error_message>.*?)\)', alert.text)
        data = match_patterns(patterns, alert.text)
        data["action"] = "error"
        if match:
            data.update(match.groupdict())
        return data
    elif alert.rule_id in ["31302", "31303"]:
        match = re.search(r'\[(warn|crit|error)\].*?: (.*?),', alert.text)
        data = match_patterns(patterns, alert.text)
        
        if match:
            data["action"] = match.group(1)
            emsg = match.group(2)
            if emsg[0] == "*":
                emsg = emsg[emsg.find(' ')+1:]
            data["error_message"] = emsg
        return data
    elif alert.rule_id == "551":
        match = re.search(r"Integrity checksum changed for: '(\S+)'", alert.text)
        if match:
            return dict(filename=match.group(1), action="changed")
    elif alert.rule_id == "554":
        match = re.search(r"New file '(\S+)' added", alert.text)
        if match:
            return dict(filename=match.group(1), action="added")
    elif alert.rule_id == "5402":
        match = re.search(r'(?P<username>[\w-]+) : PWD=(?P<pwd>\S+) ; USER=(?P<user>\w+) ; COMMAND=(?P<command>.+)', alert.text)
        if match:
            return match.groupdict()
        match = re.search(r'(?P<username>[\w-]+) : TTY=(?P<tty>\S+) ; PWD=(?P<pwd>\S+) ; USER=(?P<user>\w+) ; COMMAND=(?P<command>.+)', alert.text)
        if match:
            return match.groupdict()
    elif alert.rule_id in ["5501", "5502"]:
        match = re.search(r"session (?P<action>\S+) for user (?P<username>\S+)*", alert.text)
        if match:
            return match.groupdict()
    elif alert.rule_id in ["5704", "5705"]:
        match = re.search(r"(?P<src_ip>\d{1,3}(?:\.\d{1,3}){3}) port (?P<src_port>\d+)", alert.text)
        if match:
            return match.groupdict()
    elif alert.rule_id == "5715":
        match = re.search(r'Accepted publickey for (?P<username>\S+) from (?P<src_ip>\d+\.\d+\.\d+\.\d+) .*: (?P<ssh_key_type>\S+) (?P<ssh_signature>\S+)', alert.text)
        if match:
            return match.groupdict()
    elif alert.rule_id == "2932":
        match = re.search(r"Installed: (\S+)", alert.text)
        if match:
            return dict(package=match.group(1))
    return match_patterns(patterns, alert.text)


def match_patterns(patterns, text):
    # Search for matches in the text
    return {key: pattern.search(text).group(1) for key, pattern in patterns.items() if pattern.search(text)}


def parse_alert_json(data):
    try:
        if isinstance(data, str):
            data = objict.fromJSON(removeNonAscii(data.replace('\n', '\\n')))
    except Exception:
        data = objict.fromJSON(removeNonAscii(data))
    for key in data:
        data[key] = data[key].strip()
    if data.text:
        data.text = removeNonAscii(data.text)
    return data


def ignore_alert(alert):
    if alert.rule_id in IGNORE_RULES:
        return True
    if alert.rule_id == "510" and "/dev/.mount/utab" in alert.text:
        return True
    return False


def parse_alert_id(details):
    match = re.search(r"Alert (\d+\.\d+):", details)
    if match:
        return match.group(1)
    return ""


def parse_rule_details(details):
    alert_id = parse_alert_id(details)
    rule_pattern = r"Rule: (\d+) \(level (\d+)\) -> '([^']+)'"
    match = re.search(rule_pattern, details)
    if match:
        return objict(
            rid=int(match.group(1)), level=int(match.group(2)), 
            title=match.group(3), alert_id=alert_id)
    return objict(alert_id=alert_id)


def parse_when(alert):
    return datetime.utcfromtimestamp(int(alert.alert_id[:alert.alert_id.find(".")]))


def truncate_str(text, length):
    if len(text) > length:
        text = text[:length]
        text = text[:text.rfind(' ')] + "..."
    return text


def update_by_rule(data, geoip=None):
    if data.rule_id == "2501":
        data.title = f"SSH Auth Attempt {data.username}@{data.hostname} from {data.src_ip}"
    elif data.rule_id == "2503" and data.src_ip:
        data.title = f"SSH Auth Blocked from {data.src_ip}"
    elif data.rule_id == "31101" and data.http_status:
        data.title = f"Web {data.http_status} {data.http_method} {data.http_url} from {data.src_ip}"
    elif data.rule_id == "31104" and data.http_status:
        data.title = f"Web Attack {data.http_status} {data.http_method} {data.http_url} from {data.src_ip}"
    elif data.rule_id == "31111" and data.http_status:
        if geoip and geoip.isp:
            data.title = f"No referrer for .js - {data.http_status} {data.http_method} {data.http_url} from {data.src_ip}({geoip.isp})"
        else:
            data.title = f"No referrer for .js - {data.http_status} {data.http_method} {data.http_url} from {data.src_ip}"
    elif data.rule_id in ["31151", "31152", "31153"] and data.http_status:
        url = truncate_str(data.http_url, 50)
        data.title = f"Suspected Web Scan {url} from {data.src_ip}"
    elif data.rule_id == "31120" and data.http_status:
        url = truncate_str(data.http_url, 50)
        data.title = f"Web Error {data.http_status} {data.http_method} {url} from {data.src_ip}"
    elif data.rule_id.startswith("311") and data.http_status:
        url = truncate_str(data.http_url, 50)
        data.title = f"Web {data.http_status} {data.http_method} {url} from {data.src_ip}"
    elif data.rule_id in ["31301", "31302", "31303"] and data.error_message:
        if data.upstream and "ws/events" in data.upstream:
            data.title = f"Websocket Error: {data.error_message} on {data.hostname}"
        else:
            emsg = truncate_str(data.error_message, 50)
            data.title = f"Nginx {data.action}: {emsg} from {data.src_ip}"
    elif data.rule_id == "31516" and data.http_url:
        url = truncate_str(data.http_url, 50)
        data.title = f"Web Suspicious {data.http_status} {data.http_method} {url} from {data.src_ip}"
    elif data.rule_id == "533":
        data.title = f"Network Open Port Change Detected on {data.hostname}"
    elif data.rule_id == "5402":
        cmd = truncate_str(data.command, 50)
        data.title = f"Sudo(user: {data.user}) executed '{cmd}' on {data.hostname}"
    elif data.rule_id in ["551", "554"] and data.filename:
        name = truncate_str(data.filename, 50)
        data.title = f"File {data.action.capitalize()} on {data.hostname}: {name}"
    elif data.rule_id in ["5501", "5502"]:
        if "sudo" in data.text:
            data.title = f"Server Login {data.action} via sudo on {data.hostname}"
        else:
            data.title = f"Server Login {data.action} on {data.hostname}"
    elif data.rule_id == "5715":
        data.title = f"SSH Login Detected: {data.username}@{data.hostname} from {data.src_ip}"
    elif data.rule_id == "2932" and data.package:
        package = truncate_str(data.package, 60)
        data.title = f"Package Installed on {data.hostname}: {package}"
    elif data.src_ip and data.src_ip not in data.title:
        data.title = f"{data.title} Source IP: {data.src_ip}"
    if len(data.title) > 199:
        data.title = data.title[:199]

def parse_incoming_alert(data):
    alert = parse_alert_json(data)
    if ignore_alert(alert):
        return None
    alert.update(parse_rule_details(alert.text))
    if alert.title is None:
        return None
    alert.update(parse_alert_metadata(alert))
    if alert.src_ip in ["-", None] and alert.client:
        alert.src_ip = alert.client
    if alert.ext_ip in ["-", None]:
        alert.ext_ip = alert.src_ip
    update_by_rule(alert)
    return alert


def parseAlert(request, data):
    pdata = parse_incoming_alert(data)
    if pdata is None:
        return None
    field_names = am.ServerOssecAlert.get_model_field_names()
    soa = am.ServerOssecAlert(**{key:value for key, value in pdata.items() if key in field_names})
    soa.when = parse_when(pdata)
    soa.metadata = pdata
    if soa.src_ip is not None and len(soa.src_ip) > 6 and soa.src_ip != "127.0.0.1":
        soa.geoip = GeoIP.lookup(soa.src_ip)
        update_by_rule(pdata, soa.geoip)
        soa.title = pdata.title
    soa.save()
    return soa






# Generated by Django 4.1.5 on 2024-08-14 19:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('medialib', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='mediaitemrendition',
            name='bytes',
            field=models.BigIntegerField(help_text='Size of rendition file'),
        ),
    ]



# Generated by Django 4.2.11 on 2024-06-18 17:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0021_alter_cloudcredentials_group'),
    ]

    operations = [
        migrations.AlterField(
            model_name='memberdevice',
            name='modified',
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
    ]



from django.conf import settings
from objict import objict
import phonenumbers

try:
    import twilio
    import twilio.rest
except ImportError:
    pass

client = twilio.rest.Client(settings.TWILIO_SID, settings.TWILIO_AUTH_TOKEN)


def lookup(number, country="US", name_lookup=True):
    extra_info = ['carrier']
    if name_lookup:
        extra_info.append("caller-name")
    number = normalize(number, country)
    output = objict(success=False)
    try:
        resp = client.lookups.v1.phone_numbers(number).fetch(type=extra_info)
    except Exception as err:
        output.reason = str(err)
        return output
    output.success = True
    output.number = number
    output.carrier = objict.fromdict(resp.carrier)
    if resp.caller_name:
        output.owner_name = resp.caller_name.get("caller_name", None)
        output.owner_kind = resp.caller_name.get("caller_type", None)
    return output


def lookup2(number, country="US", name_lookup=True):
    extra_info = ['line_type_intelligence']
    if name_lookup:
        extra_info = ["caller_name"]
        # extra_info.append("sim_swap")
        # extra_info.append("call_forwarding")
        # extra_info.append("live_activity")
    number = normalize(number, country)
    output = objict(success=False)
    try:
        resp = client.lookups.v2.phone_numbers(number).fetch(fields=extra_info)
    except Exception as err:
        output.reason = str(err)
        return output
    return resp
    # output.success = True
    # output.number = number
    # output.carrier = objict.fromdict(resp.lineTypeIntelligence)
    # if resp.callerName:
    #     output.owner_name = resp.callerName.get("caller_name", None)
    #     output.owner_kind = resp.callerName.get("caller_type", None)
    # return output


    # ? access_key = 66b9eeb5c801549b999e91dab5a9e2d6
    # & number = 6172853630
    # & country_code = US
    # & format = 1

import requests
numverify_key = "66b9eeb5c801549b999e91dab5a9e2d6"
numverify_url = "http://apilayer.net/api/validate"


def verify(number, country_code="US", format=1):
    r = requests.get(numverify_url, dict(
        number=number, country_code=country_code, 
        format=format, access_key=numverify_key))
    try:
        return objict.fromdict(r.json())
    except Exception:
        pass
    return None


def validate(number, country_code="US"):
    """
    {
        'NumberValidateResponse': {
            'Carrier': 'string',
            'City': 'string',
            'CleansedPhoneNumberE164': 'string',
            'CleansedPhoneNumberNational': 'string',
            'Country': 'string',
            'CountryCodeIso2': 'string',
            'CountryCodeNumeric': 'string',
            'County': 'string',
            'OriginalCountryCodeIso2': 'string',
            'OriginalPhoneNumber': 'string',
            'PhoneType': 'string',
            'PhoneTypeCode': 123,
            'Timezone': 'string',
            'ZipCode': 'string'
        }
    }
    """
    from metrics.providers import aws
    client = aws.getClient(service="pinpoint")
    pr = client.phone_number_validate(
        NumberValidateRequest=dict(
            IsoCountryCode="US",
            PhoneNumber="+12024132409"))
    if pr:
        pr = objict.fromdict(pr)
        if pr.NumberValidateResponse:
            info = pr.NumberValidateResponse
            out = objict()
            for k, v in info:
                kl = ''.join(['_' + i.lower() if i.isupper() else i for i in k]).lstrip('_')
                out[kl] = v
            return out
    return None


def sendSMS(to_num, from_num, msg, country="US"):
    to_num = normalize(to_num, country)
    from_num = normalize(from_num, country)
    tmsg = client.messages.create(body=msg, to=to_num, from_=from_num)
    return tmsg


def sendWhatsApp(to_num, from_num, msg, country="US"):
    to_num = normalize(to_num, country)
    to_num = f"whatsapp:{to_num}"
    from_num = normalize(from_num, country)
    from_num = f"whatsapp:{from_num}"
    tmsg = client.messages.create(body=msg, to=to_num, from_=from_num)
    return tmsg


def find(text, country="US"):
    # finds any phone numbers in the text blob
    numbers = []
    for match in phonenumbers.PhoneNumberMatcher(text, country):
        numbers.append(normalize(match.number, country))
    return numbers


def isValid(number, country="US"):
    return phonenumbers.is_valid_number(phonenumbers.parse(number, country))


def normalize(raw_phone, country="US"):
    # turns a number like 202-413-2409 into +12024132409
    return convert_to_e164(raw_phone, country)


def convert_to_e164(raw_phone, country="US"):
    if not raw_phone:
        return

    if raw_phone[0] == '+':
        # Phone number may already be in E.164 format.
        parse_type = None
    else:
        # If no country code information present, assume it's a US number
        parse_type = country

    phone_representation = phonenumbers.parse(raw_phone, parse_type)
    return phonenumbers.format_number(
        phone_representation,
        phonenumbers.PhoneNumberFormat.E164)



# Copyright (C) 2010 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OAuth 2.0 utilities for Django.

Utilities for using OAuth 2.0 in conjunction with
the Django datastore.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'

import oauth2client
import base64
import pickle

from django.db import models
from oauth2client.client import Storage as BaseStorage

class CredentialsField(models.Field):
  def __init__(self, *args, **kwargs):
    if 'null' not in kwargs:
      kwargs['null'] = True
    super(CredentialsField, self).__init__(*args, **kwargs)

  def get_internal_type(self):
    return "TextField"

  def to_python(self, value):
    if value is None:
      return None
    if isinstance(value, oauth2client.client.Credentials):
      return value
    return pickle.loads(base64.b64decode(value))

  def get_db_prep_value(self, value, connection, prepared=False):
    if value is None:
      return None
    return base64.b64encode(pickle.dumps(value))


class FlowField(models.Field):
  def __init__(self, *args, **kwargs):
    if 'null' not in kwargs:
      kwargs['null'] = True
    super(FlowField, self).__init__(*args, **kwargs)

  def get_internal_type(self):
    return "TextField"

  def to_python(self, value):
    if value is None:
      return None
    if isinstance(value, oauth2client.client.Flow):
      return value
    return pickle.loads(base64.b64decode(value))

  def get_db_prep_value(self, value, connection, prepared=False):
    if value is None:
      return None
    return base64.b64encode(pickle.dumps(value))


class Storage(BaseStorage):
  """Store and retrieve a single credential to and from
  the datastore.

  This Storage helper presumes the Credentials
  have been stored as a CredenialsField
  on a db model class.
  """

  def __init__(self, model_class, key_name, key_value, property_name):
    """Constructor for Storage.

    Args:
      model: db.Model, model class
      key_name: string, key name for the entity that has the credentials
      key_value: string, key value for the entity that has the credentials
      property_name: string, name of the property that is an CredentialsProperty
    """
    self.model_class = model_class
    self.key_name = key_name
    self.key_value = key_value
    self.property_name = property_name

  def locked_get(self):
    """Retrieve Credential from datastore.

    Returns:
      oauth2client.Credentials
    """
    credential = None

    query = {self.key_name: self.key_value}
    entities = self.model_class.objects.filter(**query)
    if len(entities) > 0:
      credential = getattr(entities[0], self.property_name)
      if credential and hasattr(credential, 'set_store'):
        credential.set_store(self)
    return credential

  def locked_put(self, credentials):
    """Write a Credentials to the datastore.

    Args:
      credentials: Credentials, the credentials to store.
    """
    args = {self.key_name: self.key_value}
    entity = self.model_class(**args)
    setattr(entity, self.property_name, credentials)
    entity.save()

  def locked_delete(self):
    """Delete Credentials from the datastore."""

    query = {self.key_name: self.key_value}
    entities = self.model_class.objects.filter(**query).delete()



from . import views

import version
from . import joke
from . import helpers

import django
from django.conf import settings


from .decorators import url
from rest.extra import hostinfo
from ws4redis import client as redis
from django.db import connections


URL_PREFIX = ""

SOFTWARE_VERSIONS = getattr(settings, 'SOFTWARE_VERSIONS', None)
ALLOW_PUBLIC_SYS_INFO = getattr(settings, 'ALLOW_PUBLIC_SYS_INFO', False)
SYS_INFO_KEY = getattr(settings, 'SYS_INFO_KEY', None)
# SOFTWARE_VERSIONS_ACTUAL = {}


@url('version')
def on_get_version(request):
    return views.restStatus(request, True, {"data": version.VERSION, "ip": request.ip})


@url('versions')
def on_get_version(request):
    from rest import __version__ as restit_version
    versions = dict(
        project=version.VERSION,
        restit=restit_version)
    if request.DATA.get("detailed") and SOFTWARE_VERSIONS:
        hostinfo.getVersions(versions)
    return views.restStatus(request, True, {"data": versions})


@url('test/session')
def on_get_my_session(request):
    echo = request.DATA.get("echo", None)
    if echo:
        request.session["echo"] = echo
    return views.restReturn(request, dict(echo=request.session.get("echo", "not set")))


@url('joke')
def on_get_joke(request):
    return views.restGet(request, {"joke": joke.getRandomJoke()})


@url('status')
def on_get_status(request):
    flag = True
    try:
        flag = redis.ping()
    except Exception:
        flag = False
    status = dict(redis=flag, database={}, pool=redis.getPoolStatus())
    for alias in settings.DATABASES.keys():
        db_conn = connections[alias]
        try:
            # Ensure connection with a timeout
            db_conn.ensure_connection()
            status["database"][alias] = True
        except Exception:
            flag = False
            status["database"][alias] = False
    if flag:
        return views.restGet(request, status)
    return views.restResult(request, status, 417)

@url('system/info')
def on_get_system_info(request):
    key = request.DATA.get("key")
    if SYS_INFO_KEY is None or key != SYS_INFO_KEY:
        if not ALLOW_PUBLIC_SYS_INFO and not request.user.is_authenticated:
            return views.restPermissionDenied(request, "system info is not public", 565)
    out = hostinfo.getHostInfo(
        include_versions=request.DATA.get("versions") and SOFTWARE_VERSIONS,
        include_blocked=request.DATA.get("blocked"))
    out["redis_pool"] = redis.getPoolStatus()
    return views.restGet(request, out)


@url('system/cons')
def on_get_system_cons(request):
    key = request.DATA.get("key")
    if SYS_INFO_KEY is None or key != SYS_INFO_KEY:
        if not ALLOW_PUBLIC_SYS_INFO and not request.user.is_authenticated:
            return views.restPermissionDenied(request, "system info is not public", 565)
    cons = hostinfo.getTcpEstablished(filter=request.DATA.get("filter"))
    return views.restList(request, hostinfo.consToDict(cons))




from rest import decorators as rd
from rest import views as rv
from account.models import UserPassKey
from account import passkeys
from account.rpc.auth import on_complete_jwt


@rd.url('passkey')
@rd.url('passkey/<int:pk>')
def rest_on_processor(request, pk=None):
    return UserPassKey.on_rest_request(request, pk)


@rd.url('passkeys/register/begin')
@rd.login_required
def rest_on_passkeys_reg_begin(request):
    data, state, rp = passkeys.registerBegin(
        request.member, request,
        request.DATA.get("authenticator", "platform"))
    # state needs to be stored in a session key of some kind
    # or in redis with expires in 5m?
    request.session["fido2_state"] = state
    request.session["fido2_rp_id"] = rp.id
    return rv.restResult(request, data)


@rd.url('passkeys/register/end')
@rd.login_required
def rest_on_passkeys_reg_complete(request):
    uk = passkeys.registerComplete(
        request,
        request.session.pop("fido2_state"),
        request.session.pop("fido2_rp_id"))
    return rv.restStatus(request, True)


@rd.url('passkeys/auth/begin')
def rest_on_passkeys_auth_begin(request):
    data, state, rp = passkeys.authBegin(request)
    # state needs to be stored in a session key of some kind
    # or in redis with expires in 5m?
    request.session["fido2_state"] = state
    request.session["fido2_rp_id"] = rp.id
    return rv.restResult(request, dict(data))


@rd.url('passkeys/auth/confirm')
def rest_on_passkeys_auth_complete(request):
    uk = passkeys.authComplete(
        request,
        request.session.pop("fido2_state"),
        request.session.pop("fido2_rp_id"))
    # we now want to handle the JWT or basic login flow
    return on_complete_jwt(request, uk.member)



#!/usr/bin/env python
#
# Copyright 2010 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Common utility library."""

__author__ = ['rafek@google.com (Rafe Kaplan)',
              'guido@google.com (Guido van Rossum)',
]
__all__ = [
  'positional',
  'POSITIONAL_WARNING',
  'POSITIONAL_EXCEPTION',
  'POSITIONAL_IGNORE',
]

import inspect
import logging
import types
import urllib.request, urllib.parse, urllib.error
import urllib.parse

try:
  from urllib.parse import parse_qsl
except ImportError:
  from cgi import parse_qsl

logger = logging.getLogger(__name__)

POSITIONAL_WARNING = 'WARNING'
POSITIONAL_EXCEPTION = 'EXCEPTION'
POSITIONAL_IGNORE = 'IGNORE'
POSITIONAL_SET = frozenset([POSITIONAL_WARNING, POSITIONAL_EXCEPTION,
                            POSITIONAL_IGNORE])

positional_parameters_enforcement = POSITIONAL_WARNING

def positional(max_positional_args):
  """A decorator to declare that only the first N arguments my be positional.

  This decorator makes it easy to support Python 3 style key-word only
  parameters. For example, in Python 3 it is possible to write:

    def fn(pos1, *, kwonly1=None, kwonly1=None):
      ...

  All named parameters after * must be a keyword:

    fn(10, 'kw1', 'kw2')  # Raises exception.
    fn(10, kwonly1='kw1')  # Ok.

  Example:
    To define a function like above, do:

      @positional(1)
      def fn(pos1, kwonly1=None, kwonly2=None):
        ...

    If no default value is provided to a keyword argument, it becomes a required
    keyword argument:

      @positional(0)
      def fn(required_kw):
        ...

    This must be called with the keyword parameter:

      fn()  # Raises exception.
      fn(10)  # Raises exception.
      fn(required_kw=10)  # Ok.

    When defining instance or class methods always remember to account for
    'self' and 'cls':

      class MyClass(object):

        @positional(2)
        def my_method(self, pos1, kwonly1=None):
          ...

        @classmethod
        @positional(2)
        def my_method(cls, pos1, kwonly1=None):
          ...

  The positional decorator behavior is controlled by
  util.positional_parameters_enforcement, which may be set to
  POSITIONAL_EXCEPTION, POSITIONAL_WARNING or POSITIONAL_IGNORE to raise an
  exception, log a warning, or do nothing, respectively, if a declaration is
  violated.

  Args:
    max_positional_arguments: Maximum number of positional arguments. All
      parameters after the this index must be keyword only.

  Returns:
    A decorator that prevents using arguments after max_positional_args from
    being used as positional parameters.

  Raises:
    TypeError if a key-word only argument is provided as a positional
    parameter, but only if util.positional_parameters_enforcement is set to
    POSITIONAL_EXCEPTION.
  """
  def positional_decorator(wrapped):
    def positional_wrapper(*args, **kwargs):
      if len(args) > max_positional_args:
        plural_s = ''
        if max_positional_args != 1:
          plural_s = 's'
        message = '%s() takes at most %d positional argument%s (%d given)' % (
            wrapped.__name__, max_positional_args, plural_s, len(args))
        if positional_parameters_enforcement == POSITIONAL_EXCEPTION:
          raise TypeError(message)
        elif positional_parameters_enforcement == POSITIONAL_WARNING:
          logger.warning(message)
        else: # IGNORE
          pass
      return wrapped(*args, **kwargs)
    return positional_wrapper

  if isinstance(max_positional_args, int):
    return positional_decorator
  else:
    args, _, _, defaults = inspect.getargspec(max_positional_args)
    return positional(len(args) - len(defaults))(max_positional_args)


def scopes_to_string(scopes):
  """Converts scope value to a string.

  If scopes is a string then it is simply passed through. If scopes is an
  iterable then a string is returned that is all the individual scopes
  concatenated with spaces.

  Args:
    scopes: string or iterable of strings, the scopes.

  Returns:
    The scopes formatted as a single string.
  """
  if isinstance(scopes, str):
    return scopes
  else:
    return ' '.join(scopes)


def dict_to_tuple_key(dictionary):
  """Converts a dictionary to a tuple that can be used as an immutable key.

  The resulting key is always sorted so that logically equivalent dictionaries
  always produce an identical tuple for a key.

  Args:
    dictionary: the dictionary to use as the key.

  Returns:
    A tuple representing the dictionary in it's naturally sorted ordering.
  """
  return tuple(sorted(dictionary.items()))


def _add_query_parameter(url, name, value):
  """Adds a query parameter to a url.

  Replaces the current value if it already exists in the URL.

  Args:
    url: string, url to add the query parameter to.
    name: string, query parameter name.
    value: string, query parameter value.

  Returns:
    Updated query parameter. Does not update the url if value is None.
  """
  if value is None:
    return url
  else:
    parsed = list(urllib.parse.urlparse(url))
    q = dict(parse_qsl(parsed[4]))
    q[name] = value
    parsed[4] = urllib.parse.urlencode(q)
    return urllib.parse.urlunparse(parsed)



from .mistune.highlight import HighlightMixin
from .mistune.toc import TocMixin
from .mistune.media import MediaMixin
import mistune
try:
    HTMLRenderer = mistune.HTMLRenderer
except Exception:
    HTMLRenderer = mistune.Renderer


class WikiRenderer(TocMixin, HighlightMixin, MediaMixin, HTMLRenderer):
    def __init__(self, *args, **kwargs):
        # self.enable_math()
        self.reset_toc()
        super(WikiRenderer, self).__init__(*args, **kwargs)



# Generated by Django 4.1.4 on 2023-01-23 05:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incident', '0007_event_metadata'),
    ]

    operations = [
        migrations.AddField(
            model_name='incident',
            name='action_sent',
            field=models.DateTimeField(default=None, null=True),
        ),
        migrations.AddField(
            model_name='incident',
            name='hostname',
            field=models.CharField(db_index=True, default=None, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name='rule',
            name='action_after',
            field=models.IntegerField(default=0),
        ),
        migrations.AddField(
            model_name='rule',
            name='bundle_by',
            field=models.IntegerField(default=3),
        ),
    ]



from django.contrib import admin
from .models import Task, TaskHook
from django import forms


class TaskAdmin(admin.ModelAdmin):
    pass


admin.site.register(Task, TaskAdmin)
admin.site.register(TaskHook)



from sessionlog.models import SessionLog

from django.db import models
from django.conf import settings

import copy

from rest.models import RestModel
from rest import helpers
from rest.log import getLogger
from auditlog import cloudwatch

import traceback

AUDIT_LOGGER = getLogger("auditlog", filename="auditlog.log")
AUDITLOG_LOGGER_BY_USER = getattr(settings, "AUDITLOG_LOGGER_BY_USER", False)
AUDITLOG_LOGGERS = getattr(settings, "AUDITLOG_LOGGERS", {})

PERSISTENT_LOG_PRINT = getattr(settings, "PERSISTENT_LOG_PRINT", False)


MAX_AUDIT_LINES = 400
MAX_AUDIT_LENGTH = 9000
MAX_AUDIT_LINE_LENGTH = 200


class ConsoleColors:
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    PINK = "\033[35m"
    BLUE = '\033[34m'
    WHITE = '\033[37m'

    HBLACK = '\033[90m'
    HRED = '\033[91m'
    HGREEN = '\033[92m'
    HYELLOW = '\033[93m'
    HBLUE = '\033[94m'
    HPINK = "\033[95m"
    HWHITE = '\033[97m'

    HEADER = '\033[95m'
    FAIL = '\033[91m'
    OFF = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


LOG_FILTERS = getattr(settings, "AUDIT_LOG_FILTERS", {})


class PersistentLog(models.Model, RestModel):
    class RestMeta:
        DEFAULT_SORT = "-when"
        CAN_SAVE = False
        GROUP_FIELD = None  # do this so when we sort by group it is always exact
        QUERY_FIELDS = [
            "level", "component", "pkey", "action", 
            "request_method", "request_path", 
            "session__ip", "user",
            "tid", "group", "user__username"]
        SEARCH_FIELDS = ["user__username", "user__last_name", "message", "session__ip", "request_path"]
        SEARCH_TERMS = [
            "message", "action", "component", "tid",
            "pkey", ("component_id", "pkey"),
            ("ip", "session__ip"),
            ("path", "request_path"),
            ("method"), "request_method",
            ("username", "user__username"),
            ("group", "group__pk"),
            ("country", "session__location__country"),
            ("state", "session__location__state"),
            ("country", "session__location__country")]
        DATE_RANGE_FIELD = "when"
        # number of days to go back (increase performance)
        DATE_RANGE_DEFAULT = 7

        FORMATS = {
            "csv": [
                "when",
                "level",
                "request_method",
                "request_path",
                "action",
                "message",
                "session.ip",
                "user.username",
                "group.name",
                "tid"
            ]
        }

        GRAPHS = {
            "default": {
                "graphs": {
                    "user": "default",
                    "group": "default",
                    "session": "default"
                }
            },
            "list": {
                "graphs": {
                    "user": "default",
                    "group": "default",
                    "session": "default"
                }
            }
        }

    when = models.DateTimeField(db_index=True, auto_now_add=True, editable=False, help_text="When change was logged")
    level = models.IntegerField(default=0)
    message = models.TextField(blank=True, null=True, default=None)

    action = models.TextField(max_length=127, blank=True, null=True, default=None, db_index=True)
    component = models.CharField(max_length=127, blank=True, null=True, default=None, db_index=True)
    pkey = models.IntegerField(blank=True, null=True, default=None, db_index=True)
    session = models.ForeignKey(SessionLog, null=True, blank=True, on_delete=models.SET_NULL)
    request_method = models.CharField(max_length=254, blank=True, null=True, default=None)
    tid = models.CharField(max_length=254, blank=True, null=True, default=None, db_index=True)
    request_path = models.TextField(blank=True, null=True, default=None)
    user = models.ForeignKey("account.User", null=True, default=None, on_delete=models.CASCADE)
    group = models.ForeignKey("account.Group", null=True, default=None, on_delete=models.CASCADE)

    def setMessage(self, message, truncate=True, sanatize=True):
        if message is None:
            self.message = "None"
            return
        if isinstance(message, dict) or type(message) is list:
            if sanatize and isinstance(message, dict):
                message = PersistentLog.sanatize(message)
            self.message = helpers.dictToString(message, no_truncate=not truncate)
        else:
            self.message = helpers.toString(message)

        if truncate:
            if self.message.count('\n') > MAX_AUDIT_LINES:
                p = helpers.find_nth(self.message, '\n', MAX_AUDIT_LINES)
                self.message = self.message[:p]
                self.message += "\n ... <truncated lines>"

            if len(self.message) > MAX_AUDIT_LENGTH:
                lines = self.message.split('\n')
                tlines = []
                for ln in lines:
                    if len(ln) >= MAX_AUDIT_LINE_LENGTH:
                        tlines.append(ln[:MAX_AUDIT_LINE_LENGTH] + '...<truncated line len>')
                    else:
                        tlines.append(ln)
                self.message = '\n'.join(tlines)

        # check for invalid characters
        if self.message.find("\0") >= 0:
            self.message = self.message.replace("\0", "<null>")
            # helpers.log_error("found null characters", self.message)

    def printToLog(self):
        print_log = ["{}{}:{}{}".format(
            ConsoleColors.YELLOW, self.request_path, self.request_method, ConsoleColors.OFF)]
        if self.action == "request":
            print_log.append("{0}------- BEGIN --------------".format(ConsoleColors.YELLOW))
        print_log.append('{0}***** BEGIN {1}.{2} *****{3}'.format(
            ConsoleColors.YELLOW, self.component, self.action, ConsoleColors.HGREEN))
        if self.user and self.user.is_authenticated:
            print_log.append("USER: {}".format(self.user.username))
        if self.group:
            print_log.append("GROUP: {}".format(self.group.name))
        if self.tid:
            print_log.append("TID: {}".format(self.tid))
        if self.message:
            print_log.append(helpers.toString(self.message))
        else:
            print_log.append("no data")
        print_log.append('{0}***** END {1}.{2} *****{3}'.format(
            ConsoleColors.YELLOW, self.component, self.action, ConsoleColors.OFF))
        if self.action == "response":
            print_log.append("{}------- END {}:{}  --------------{}\n".format(
                ConsoleColors.YELLOW, self.request_path, self.request_method, ConsoleColors.OFF))
        if print_log:
            helpers.log_print(*print_log)

    @staticmethod
    def sanatize_password(value, n=1):
        if value is None or not isinstance(value, str):
            return value
        st = value
        return ''.join("*" if i % n == 0 else char for i, char in enumerate(st, 1))

    @staticmethod
    def sanatize_track(value):
        if value is None or not isinstance(value, str):
            return value
        st = value
        if len(st) > 12:
            pan = None
            pan2 = None
            parts = st.split('?')
            for p1 in parts:
                p1 = p1.upper().replace(';', '').replace('%B', '').replace('B', '')
                ex_data = None
                if '=' in p1:
                    pos = p1.find('=')
                    pan = p1[:pos]
                    ex_data = p1[pos + 8:]
                    st = st.replace(ex_data, "*" * len(ex_data))
                elif '^' in p1:
                    pan = p1[:p1.find('^')]
                    pos = p1.rfind('^')
                    ex_data = p1[pos + 8:]
                    st = st.replace(ex_data, "*" * len(ex_data))
                elif "D2" in p1:
                    p1 = p1.replace("D2", "=")
                    pos = p1.find('=')
                    pan = p1[:pos]
                    ex_data = p1[pos + 8:]
                    st = st.replace(ex_data, "*" * len(ex_data))
                else:
                    continue
                if not pan2:
                    pan2 = pan[:6] + "***" + pan[-4:]
            if pan is not None:
                return st.replace(pan, pan2)
        return value

    @staticmethod
    def sanatize_pan(value):
        if value and len(value) > 10:
            return "{}***{}".format(value[:6], value[-4:])
        return value

    @staticmethod
    def sanatize_epb(value):
        if isinstance(value, str):
            return "{}:*****".format(len(value))
        return value

    @staticmethod
    def sanatize_ssn(value):
        if value:
            return "***-**-{}".format(value[-4:])
        return value

    @staticmethod
    def sanatize_all(value):
        if isinstance(value, str):
            return "*" * len(value)
        return value

    @staticmethod
    def sanatize(obj):
        # deepcopy vs copy?
        sobj = copy.copy(obj)
        for key in sobj:
            if key is None:
                continue
            lkey = key.lower()
            if lkey in LOG_FILTERS:
                funcn = LOG_FILTERS.get(lkey)
                if callable(funcn):
                    sobj[key] = funcn(sobj[key])
                elif hasattr(PersistentLog, funcn):
                    sobj[key] = getattr(PersistentLog, funcn)(sobj[key])
            elif lkey in ["pan", "cardnumber"]:
                sobj[key] = PersistentLog.sanatize_pan(sobj[key])
            elif "key" in lkey or lkey in ["5a", "57", "99"]:
                sobj[key] = PersistentLog.sanatize_all(sobj[key])
            elif lkey.startswith("track"):
                sobj[key] = PersistentLog.sanatize_track(sobj[key])
            elif "password" in lkey:
                sobj[key] = PersistentLog.sanatize_password(sobj[key])
            elif lkey == "epb":
                sobj[key] = PersistentLog.sanatize_epb(sobj[key])
            elif lkey == "ssn":
                sobj[key] = PersistentLog.sanatize_ssn(sobj[key])
            # lets truncate
            if type(sobj[key]) in [str, str] and len(sobj[key]) > MAX_AUDIT_LINE_LENGTH:
                sobj[key] = sobj[key][:MAX_AUDIT_LINE_LENGTH] + '...<truncated linelen>'
            if isinstance(sobj[key], dict):
                sobj[key] = PersistentLog.sanatize(sobj[key])
        return sobj

    @staticmethod
    def log(message, level=0, request=None, component=None,
            pkey=None, action=None, group=None, path=None,
            method=None, tid=None, no_truncate=False,
            aws_log_group=None, aws_log_stream=None):
        plog = PersistentLog.createLogFromRequest(
            request, component=component, tid=tid, pkey=pkey, action=action, group=group,
            path=path, method=method, level=level)

        plog.setMessage(message, truncate=not no_truncate)

        try:
            plog.save()
            if aws_log_group and aws_log_stream:
                cloudwatch.log(plog.toDict(), aws_log_group, aws_log_stream)
        except Exception:
            helpers.log_exception(plog.message)

        if PERSISTENT_LOG_PRINT:
            plog.printToLog()
        return plog

    @staticmethod
    def logException(message=None, level=50, request=None, component=None,
                     pkey=None, action="error", group=None, path=None,
                     method=None, tid=None, no_truncate=True):
        # LETS LOG THE REQUEST DATA TO BE SURE
        if request is None:
            from rest.middleware import get_request
            request = get_request()

        helpers.log_exception(message)
        if request:
            helpers.log_error(request.DATA.toDict())
            # request.DATA.log()
            helpers.log_error("tid={}\npath={}\nip={}\nuser={}".format(
                tid, path, request.ip, request.user))

        # LOG THE EXCEPTION TO local LOGGERS
        plog = PersistentLog.createLogFromRequest(
            request=request, component=component, tid=tid, pkey=pkey, action=action, group=group,
            path=path, method=method, level=level)
        stack = str(traceback.format_exc())
        plog.setMessage(stack, truncate=not no_truncate)
        try:
            plog.save()
        except Exception:
            helpers.log_exception(plog.message)
        return plog

    @staticmethod
    def logError(message=None, level=0, request=None, component=None,
                 pkey=None, action="error", group=None, path=None,
                 method=None, tid=None, no_truncate=True):
        # LETS LOG THE REQUEST DATA TO BE SURE
        if request is None:
            from rest.middleware import get_request
            request = get_request()
        if request:
            request.DATA.log()
        plog = PersistentLog.createLogFromRequest(
            request, component=component, tid=tid, pkey=pkey, action=action, group=group,
            path=path, method=method, level=level)
        stack = str(traceback.format_exc())
        plog.setMessage(stack, truncate=not no_truncate)
        try:
            plog.save()
        except Exception:
            helpers.log_exception(plog.message)
        # print all errors to local logs
        plog.printToLog()
        return plog

    @staticmethod
    def createLogFromRequest(request=None, component=None, tid=None,
                             pkey=None, action=None, group=None,
                             path=None, method=None, level=10):
        if request is None:
            from rest.middleware import get_request
            request = get_request()
        if pkey is None and request is not None and hasattr(request, "_log_component"):
            if request._log_pk is not None:
                component = request._log_component
                pkey = request._log_pk
        plog = PersistentLog(
            group=group, level=level, component=component, tid=tid,
            pkey=pkey, action=action, request_path=path, request_method=method)
        if request:
            if request.user.is_authenticated:
                plog.user = request.user
            if not group and hasattr(request, "group"):
                plog.group = request.group
            if not path:
                plog.request_path = request.path
            else:
                plog.request_path = path
            if not method:
                plog.request_method = request.method
            else:
                plog.request_method = method
            plog.session = SessionLog.GetSession(request)
            if plog.tid is None:
                if hasattr(request, "device_id") and request.device_id:
                    plog.tid = request.device_id
                elif hasattr(request, "buid") and request.buid:
                    plog.tid = request.buid
        return plog


class AuditLog(models.Model):
    """
    Audit Log
    """

    class Meta:
        permissions = (
            ("can_read", "Can read audit log"),
        )

    when = models.DateTimeField(auto_now_add=True, editable=False, help_text="When change was logged")
    model = models.CharField(max_length=127, db_index=True)
    pkey = models.IntegerField(db_index=True)
    attribute = models.CharField(max_length=127)
    user = models.ForeignKey("account.User", null=True, on_delete=models.CASCADE)
    session = models.ForeignKey(SessionLog, null=True, blank=True, on_delete=models.CASCADE)
    how = models.TextField(max_length=127, null=True, blank=True)
    referer = models.TextField(null=True, blank=True)
    stack = models.TextField(null=True, blank=True)
    oldval = models.TextField(null=True, blank=True)
    newval = models.TextField(null=True, blank=True)

    def __unicode__(self):
        return "[%s] %s.%s" % (self.when, self.model, self.attribute)

    def reference(self):
        if self.referer:
            try:
                ret = self.referer.partition(settings.BASE_URL)
            except AttributeError:
                pass
            else:
                if ret[0] == '':
                    return ret[2]
            try:
                ret = self.referer.partition(settings.BASE_URL_SECURE)
            except AttributeError:
                pass
            else:
                if ret[0] == '':
                    return ret[2]

        try:
            return self.session.user_agent.partition(' ')[0]
        except (SessionLog.DoesNotExist, AttributeError):
            pass

        return None



try:
    import cv2
    import numpy as np
except:
    cv2 = None
import os
from . import contours

face_cascade = None
eye_cascade = None

def loadFaceDetection():
    global face_cascade, eye_cascade
    face_cascade = cv2.CascadeClassifier('haarcascade_frontalface_default.xml')
    eye_cascade = cv2.CascadeClassifier('haarcascade_eye.xml')

class CVImage(object):
    def __init__(self, path=None, image=None):
        self.path = path
        self.image = image
        self.last_image = None
        self.gray_img = None
        self._kernel = None
        if image is None and path:
            if not os.path.exists(path):
                raise Exception("file does not exist: {}".format(path))
            self.image = cv2.imread(self.path)
            self._org_img = self.image
        self.contours = None

    @property
    def is_grayscale(self):
        return len(self.image.shape) == 2

    @property
    def width(self):
        h, w = self.image.shape[:2]
        return w

    @property
    def height(self):
        h, w = self.image.shape[:2]
        return h 

    def reset(self):
        self.setImage(self._org_img)
        return self

    def setImage(self, image):
        self.gray_img = None
        self.last_image = self.image
        self.image = image
        return self

    def grayscale(self, preserve_orig=False):
        if self.gray_img != None:
            return self
        if self.is_grayscale:
            self.gray_img = self.image
            return self
        gray = cv2.cvtColor(self.image, cv2.COLOR_BGR2GRAY)
        if not preserve_orig:
            self.setImage(gray)
        self.gray_img = gray
        return self

    def bw(self, threshold=127, threshold_convert=255):
        return self.blackAndWhite()

    def blackAndWhite(self, threshold=127, threshold_convert=255):
        self.grayscale(True)
        thresh, bw_img = cv2.threshold(self.gray_img, threshold, threshold_convert, cv2.THRESH_BINARY)
        self.setImage(bw_img)
        return self

    def isBlurry(self, threshold=100.0):
        self.grayscale(True)
        fm = cv2.Laplacian(self.gray_img, cv2.CV_64F).var()
        if fm < threshold:
            return True
        return False

    def toB64(self, frame):
        return str(base64.b64encode(frame), encoding="utf-8")

    def copy(self):
        return CVImage(image=self.image)

    def save(self, path=None, ext="png"):
        if path is None:
            path = self.path
        filename, cext = os.path.splitext(path)
        if cext.lower() != ext:
            opath = path
            path = "{}.{}".format(path, ext)
            cv2.imwrite(path, self.image)
            os.rename(path, opath)
        else:
            cv2.imwrite(path, self.image)

    def drawBoxes(self, boxes, color=(255, 0, 0), thickness=2):
        if isinstance(boxes[0][0], tuple):
            for xy1, xy2 in boxes:
                cv2.rectangle(self.image, xy1, xy2, color, thickness)
        else:
            for x, y, w, h in boxes:
                x2 = x + w
                y2 = y + h
                cv2.rectangle(self.image, (x, y), (x2, y2), color, thickness)
        return self

    def drawContours(self, contours=None, index=-1, color=(255, 0, 0), thickness=2):
        if contours is None:
            if self.contours is None:
                contours = self.findContours()
            else:
                contours = self.contours
        for i in range(0, len(contours)):
            cv2.drawContours(self.image, contours, i, color, thickness)
        return self

    def findContours(self, method=0):
        if not self.is_grayscale:
            return self.findContoursAdv()
        cnts = cv2.findContours(self.image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        self.contours =  contours.grab_contours(cnts)
        return self.contours

    def findContoursAdv(self):
        self.grayscale(True)
        blurred = cv2.GaussianBlur(self.gray_img, (5, 5), 0)
        ret, thresh = cv2.threshold(blurred, 60, 255, cv2.THRESH_BINARY)
        cnts = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        self.contours =  contours.grab_contours(cnts)
        return self.contours

    def detectBox(self):
        self.grayscale(True)
        edges = cv2.Canny(self.gray_img, 75, 150)
        lines = cv2.HoughLinesP(edges, 1, np.pi/180, 30, maxLineGap=250)
        for line in lines:
            x1, y1, x2, y2 = line[0]
            cv2.line(self.image, (x1, y1), (x2, y2), (0, 0, 0), 5)
        return self

    def crop(self, x, y, w, h, scale_width=1.0, scale_height=1.0):
        height, width, channels = self.image.shape
        scaled_w = int(w * scale_width)
        scaled_h = int(h * scale_height)
        scaled_x = max(0, int(x - (scaled_w - w) / 2))
        scaled_y = max(0, int(y - (scaled_h - h) / 2))
        scaled_x = min(scaled_x, width)
        scaled_y = min(scaled_y, height)
        # this could go past the height of the image!!
        crop_x = scaled_x + scaled_w
        crop_y = scaled_y + scaled_h
        self.setImage(self.image[scaled_y:crop_y, scaled_x:crop_x])
        return self

    def resize(self, factor=0, width=None, height=None, inter=3):
        dim = None
        (h, w) = self.image.shape[:2]
        if factor == 0 and width is None and height is None:
            return self

        if factor:
            nh = h * factor
            nw = w * factor
            dim = (nw, nh)
        else:
            # check to see if the width is None
            if width is None:
                r = height / float(h)
                dim = (int(w * r), height)
            else:
                r = width / float(w)
                dim = (width, int(h * r))
        self.last_image = self.image
        self.setImage(cv2.resize(self.image, dim, interpolation=inter))
        return self

    def translate(self, x, y):
        # define the translation matrix and perform the translation
        M = np.float32([[1, 0, x], [0, 1, y]])
        shifted = cv2.warpAffine(self.image, M, (self.image.shape[1], self.image.shape[0]))
        self.setImage(shifted)
        # return the translated image
        return self

    def rotate(self, angle, center=None, scale=1.0):
        (h, w) = self.image.shape[:2]
        if center is None:
            center = (w // 2, h // 2)
        M = cv2.getRotationMatrix2D(center, angle, scale)
        rotated = cv2.warpAffine(self.image, M, (w, h))
        self.setImage(rotated)
        return self

    def rotateBound(self, angle):
        (h, w) = self.image.shape[:2]
        (cX, cY) = (w / 2, h / 2)
        M = cv2.getRotationMatrix2D((cX, cY), -angle, 1.0)
        cos = np.abs(M[0, 0])
        sin = np.abs(M[0, 1])
        nW = int((h * sin) + (w * cos))
        nH = int((h * cos) + (w * sin))
        M[0, 2] += (nW / 2) - cX
        M[1, 2] += (nH / 2) - cY
        self.setImage(cv2.warpAffine(self.image, M, (nW, nH)))
        return self

    def autoCanny(self, sigma=0.33):
        # compute the median of the single channel pixel intensities
        v = np.median(self.image)
        # apply automatic Canny edge detection using the computed median
        lower = int(max(0, (1.0 - sigma) * v))
        upper = int(min(255, (1.0 + sigma) * v))
        edged = cv2.Canny(self.image, lower, upper)
        self.setImage(edged)
        return self

    def adjustGamma(self, gamma=1.0):
        # build a lookup table mapping the pixel values [0, 255] to
        # their adjusted gamma values
        invGamma = 1.0 / gamma
        table = np.array([((i / 255.0) ** invGamma) * 255 for i in np.arange(0, 256)]).astype("uint8")
        # apply gamma correction using the lookup table
        self.setImage(cv2.LUT(self.image, table))
        return self

    @property
    def kernel(self):
        if self._kernel is None:
            self.setKernel()
        return self._kernel
    
    def setKernel(self, size=(5,5)):
        self._kernel = np.ones(size,np.uint8)
        return self

    def erode(self, iterations=1):
        self.setImage(cv2.erode(self.image,self.kernel,iterations))
        return self

    def dilate(self, iterations=1):
        self.setImage(cv2.dilate(self.image,self.kernel,iterations))
        return self

    def opening(self):
        self.setImage(cv2.morphologyEx(self.image,cv2.MORPH_OPEN, self.kernel))
        return self

    def closing(self):
        self.setImage(cv2.morphologyEx(self.image,cv2.MORPH_CLOSE, self.kernel))
        return self

    def laplacian(self):
        self.setImage(cv2.Laplacian(self.image,cv2.CV_64F))
        return self

    def invert(self):
        self.setImage(cv2.bitwise_not(self.image))
        return self

    def brightness(self, level=0):
        return self.adjust(level)

    def contrast(self, level=0):
        return self.adjust(0, level)

    def blur(self, size=(5,5)):
        self.setImage(cv2.blur(self.image,size))
        return self

    def gaussianBlur(self, size=(5,5), sigma_x=0, sigma_y=0):
        self.setImage(cv2.GaussianBlur(self.image,size, sigma_x, sigma_y))
        return self

    def medianBlur(self, ksize=5):
        self.setImage(cv2.medianBlur(self.image,ksize))
        return self

    def bilateralFilter(self, diameter=9, sigma_color=75, sigma_space=75):
        self.setImage(cv2.bilateralFilter(self.image,diameter,sigma_color,sigma_space))
        return self

    def adaptiveThreshold(self, max_value=255, method=1, threshold_type=0, size=11, k=2):
        self.setImage(cv2.adaptiveThreshold(self.image,max_value,method, threshold_type,size,k))
        return self

    def autoBalance(self, clip_limit=2.0, tile_grid_size=(8,8)):
        # create a CLAHE object (Arguments are optional).
        # clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=tile_grid_size)
        # self.setImage(clahe.apply(self.image))
        self.grayscale()
        self.setImage(cv2.equalizeHist(self.gray_img))
        return self
    
    def adjust(self, brightness=0.0, contrast=0.0):
        """
        Adjust the brightness and/or contrast of an image
        :param image: OpenCV BGR image
        :param contrast: Float, contrast adjustment with 0 meaning no change
        :param brightness: Float, brightness adjustment with 0 meaning no change
        """
        beta = 0
        # See the OpenCV docs for more info on the `beta` parameter to addWeighted
        # https://docs.opencv.org/3.4.2/d2/de8/group__core__array.html#gafafb2513349db3bcff51f54ee5592a19
        self.setImage(cv2.addWeighted(self.image,
                               1 + float(contrast) / 100.,
                               self.image,
                               beta,
                               float(brightness)))
        return self

    def cropAndRotate(self, rect, box, crop_width_scale=1.2, crop_height_scale=1.2):
        width = rect[1][0]
        height = rect[1][1]
        angle = rect[2]
        rotated = False

        Xs = [i[0] for i in box]
        Ys = [i[1] for i in box]
        x1 = min(Xs)
        x2 = max(Xs)
        y1 = min(Ys)
        y2 = max(Ys)

        if angle < -45:
            angle += 90
            rotated = True

        center = (int((x1 + x2) / 2), int((y1 + y2) / 2))
        size = (int(crop_width_scale * (x2 - x1)), int(crop_height_scale * (y2 - y1)))

        M = cv2.getRotationMatrix2D((size[0] / 2, size[1] / 2), angle, 1.0)

        cropped = cv2.getRectSubPix(self.image, size, center)
        cropped = cv2.warpAffine(cropped, M, size)

        cropped_width = width if not rotated else height
        cropped_height = height if not rotated else width

        cropped_rotated = cv2.getRectSubPix(
            cropped,
            (int(cropped_width * crop_width_scale), int(cropped_height * crop_height_scale)),
            (size[0] / 2, size[1] / 2),
        )
        self.setImage(cropped_rotated)
        return self

    def findLargestFace(self, faces=None):
        max_area = 0
        big_face = None
        for (x, y, w, h) in faces:
            new_area = w * h
            if new_area > max_area:
                big_face = (x, y, w, h)
        return big_face

    def detectEyes(self, face):
        (x,y,w,h) = face
        roi_gray = self.gray_img[y:y+h, x:x+w]
        eyes = eye_cascade.detectMultiScale(roi_gray)
        return eyes

    def detectFaces(self, scale_factor=1.2, min_neighbors=10):
        self.grayscale(True) 
        if face_cascade is None:
            loadFaceDetection()
        faces = face_cascade.detectMultiScale(self.gray_img, scale_factor, min_neighbors)
        self.faces = faces
        return faces



from Crypto.Cipher import AES
from Crypto import Random
import hashlib
from hashlib import md5, sha256

import json
from . import util


DEFAULT_SALT = b"Salted__"
MD5V_SALT = b"__MD5V__"
SDKV_SALT = b"__SDKV__"
SDKV16_SALT = b"__SDKV16"
SHA256_SALT = b"SALTY__"

KNOWN_SALTS = [DEFAULT_SALT, SDKV_SALT, MD5V_SALT, SDKV16_SALT, SHA256_SALT]

DEFAULT_IV_GENERATOR = "md5"


def encrypt(key, data, random_size=128, ivgen=DEFAULT_IV_GENERATOR, base64=True):
    """
    encrypts @data with @key using AES
    will pad data with random string of @random_size
        this will keep it so the same data is never producing the same encryption
    returns bytes as base64 string

    HIS USES AES/CBC/PKCS7Padding
    """
    if random_size > 0:
        rdata = util.toBytes(data) + util.toBytes(util.randomString(random_size))
    else:
        rdata = data
    if not base64:
        return cbc_encrypt(key, rdata, ivgen, base64)
    return util.toString(cbc_encrypt(key, rdata, ivgen))


def decrypt(key, edata, random_size=128, as_string=True, base64=True):
    if base64:
        encrypted = util.fromBase64(edata, False)
    else:
        encrypted = edata
    data = cbc_decrypt(key, encrypted)
    if random_size > 0:
        data = data[:-1 * random_size]
    if as_string:
        return util.toString(data)
    return data


def cbc_encrypt(key, data, ivgen=DEFAULT_IV_GENERATOR, base64=True):
    if isinstance(data, dict) or isinstance(data, list):
        data = json.dumps(data)
    salt_prefix = getSaltPrefix(ivgen)
    data = util.toBytes(data)
    key = util.toBytes(key)
    salt = Random.new().read(8)
    if ivgen == "simple": 
        key_iv = deterministric_key_iv(ivgen, key, salt, 32)
        key = key_iv[:16]
        iv = key_iv[16:]
    else:
        key_iv = deterministric_key_iv(ivgen, key, salt, 32 + 16)
        key = key_iv[:32]
        iv = key_iv[32:]
    aes = AES.new(key, AES.MODE_CBC, iv)
    if not base64:
        return salt_prefix + salt + aes.encrypt(util.pad(data))
    return util.toBase64(salt_prefix + salt + aes.encrypt(util.pad(data)))


def cbc_decrypt(key, encrypted):
    salt_prefix = encrypted[0:8]
    if salt_prefix not in KNOWN_SALTS:
        print("unknown salt: {}".format(str(salt_prefix)))
    key = util.toBytes(key)
    salt = encrypted[8:16]
    if salt_prefix == SDKV16_SALT:  
        key_iv = deterministric_key_iv(salt_prefix, key, salt, 32)
        key = key_iv[:16]
        iv = key_iv[16:]
    else:
        key_iv = deterministric_key_iv(salt_prefix, key, salt, 32 + 16)
        key = key_iv[:32]
        iv = key_iv[32:]
    aes = AES.new(key, AES.MODE_CBC, iv)
    decrypted = aes.decrypt(encrypted[16:])
    return util.unpad(decrypted)


def deterministric_key_iv(generator, data, salt, length=48):
    if generator in [DEFAULT_SALT, MD5V_SALT, "md5"]:
        return md5_deterministic_key_iv(data, salt, length)
    elif generator == SHA256_SALT:
        return sha256_deterministic_key_iv(data, salt, length)
    return simple_deterministic_key_iv(data, salt, length)


def getSaltPrefix(generator):
    if generator == "md5":
        return DEFAULT_SALT
    elif generator == "simple":
        return SDKV16_SALT
    return SDKV_SALT


def md5_deterministic_key_iv(data, salt, length=48):
    # extended from https://gist.github.com/gsakkis/4546068
    assert len(salt) == 8, len(salt)
    data += salt
    key = md5(data).digest()
    final_key = key
    while len(final_key) < length:
        key = md5(key + data).digest()
        final_key += key
    return final_key[:length]


def sha256_deterministic_key_iv(data, salt, length=48):
    # extended from https://gist.github.com/gsakkis/4546068
    assert len(salt) == 8, len(salt)
    data += salt
    key = sha256(data).digest()
    final_key = key
    while len(final_key) < length:
        key = sha256(key + data).digest()
        final_key += key
    return final_key[:length]


def simple_deterministic_key_iv(data, salt, length=48):
    src = data + salt
    src_len = len(src)
    dst = bytearray(length)
    p = salt[2] >> 4
    while p >= src_len:
        p = int(p / 2)
    x = len(data)
    for i in range(0, length):
        h = (src[p] - src[x]) * (src[p])
        dst[i] = h & 255
        p -= 1
        if p <= 0:
            p = src_len - 1
        x += 1
        if x >= src_len:
            x = 0
    return bytes(dst)








from ..engines import websnap
from ..render_utils import *
from ..exceptions import *

from PIL import Image

from django.conf import settings

PresetConfig = {
    'name':		'WebSnap',
    'description':	'Snapshot a URL',
    'stage':	101,
    'default_use':	'websnap',
    'applicable_kinds': 'E',
}



def render(item, renditiondef, params):
    fp = tempfile.NamedTemporaryFile(suffix=('.' + params.get('image_container', 'png')), prefix="t_websnap_")
    dstfile = fp.name
    tmpdir = tempfile.mkdtemp()
    prevdir = os.getcwd()
    os.chdir(tmpdir)

    original = get_rendition(item, 'original')

    websnap.run(dstfile, original.url)
    im = Image.open(dstfile)
    params["width"] = im.size[0]
    params["height"] = im.size[1]

    new_rendition(item, renditiondef, fp, kind='I', width=params.get('width', None), height=params.get('height', None))




# coding: utf-8

"""
    mistune_contrib.highlight
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Support highlight code features for mistune.

    :copyright: (c) 2014 - 2015 by Hsiaoming Yang.
"""

import mistune
import urllib.parse

def parseWidthHeight(link, w=0, h=0):
    o = urllib.parse.urlparse(link)
    if o.query:
        params = urllib.parse.parse_qs(o.query)
        if "md_width" in params:
            w = params.get("md_width")[0]
        if "md_height" in params:
            h = params.get("md_height")[0]
    return w, h

class MediaMixin(object):
    HARMFUL_PROTOCOLS = {
        'javascript:',
        'vbscript:',
        'data:',
    }
    def _safe_url(self, url):
        schemes = self.HARMFUL_PROTOCOLS
        if schemes:
            for s in schemes:
                if url.startswith(s):
                    url = '#harmful-link'
                    break
        return url

    def link(self, link, text=None, title=None):
        label = text
        if text is None:
            text = link
        o = urllib.parse.urlparse(link)
        if o.path.endswith(".mp4"):
            return self.video(link, text, title)
        href = self._safe_url(link)
        params = {"href":href}
        if title:
            params["title"] = title
        if label:
            params["download"] = label
        if not href.startswith("http"):
            params["data-action"] = "local_page"
        flat_params = ' '.join("{}='{}'".format(key,val) for (key,val) in list(params.items()))
        return """<a {}>{}</a>""".format(flat_params, (text or link))

    def video(self, link, text=None, title=None, w=0, h=0):
        w, h = parseWidthHeight(link, w, h)
        params = {}
        if title:
            params["title"] = title
        if w:
            params["width"] = w
        if h:
            params["height"] = h
        flat_params = ' '.join("{}='{}'".format(key,val) for (key,val) in list(params.items()))
        return """<video controls {}><source src="{}" type="video/mp4"></video>""".format(flat_params, link)

    def image(self, src, alt="", title=None):
        if not src:
            return ""
        o = urllib.parse.urlparse(src)
        if o.path.endswith(".mp4"):
            return self.video(src, alt, title)
        w, h = parseWidthHeight(src)
        params = {}
        if title:
            params["title"] = title
        if w:
            params["width"] = w
        if h:
            params["height"] = h
        if alt:
            params["alt"] = alt
        params["src"] = src

        flat_params = ' '.join("{}='{}'".format(key,val) for (key,val) in list(params.items()))
        return '<img {} />'.format(flat_params)




from objict import objict
from rest import settings
import requests

IPAPI_TOKEN = settings.get("GEOIP_IPAPI_TOKEN", "6551a9fe6f22a29bd4e4f176920a4646")


def getLocation(ip):
    """
    {
       "ip": "72.219.88.137",
       "type": "ipv4",
       "continent_code": "NA",
       "continent_name": "North America",
       "country_code": "US",
       "country_name": "United States",
       "region_code": "CA",
       "region_name": "California",
       "city": "Dana Point",
       "zip": "92629",
       "latitude": 33.485538482666016,
       "longitude": -117.69378662109375,
       "location":
       {
           "geoname_id": 5341483,
           "capital": "Washington D.C.",
           "languages":
           [
               {
                   "code": "en",
                   "name": "English",
                   "native": "English"
               }
           ],
           "country_flag": "https://assets.ipstack.com/flags/us.svg",
           "country_flag_emoji": "🇺🇸",
           "country_flag_emoji_unicode": "U+1F1FA U+1F1F8",
           "calling_code": "1",
           "is_eu": false
       }
    }
    """
    url = "http://api.ipapi.com/{}".format(ip)
    params = dict(access_key=IPAPI_TOKEN)
    resp = requests.get(url, params=params)
    data = objict.fromdict(resp.json())

    loc = objict(provider="ipapi.com")
    loc.isp = ""
    loc.ip = data.ip
    loc.hostname = ""
    loc.ip_type = data.type
    loc.latitude = data.latitude
    loc.longitude = data.longitude
    loc.continent = data.continent_name
    loc.region = data.region_name
    loc.country = data.country_name
    loc.countryCode = data.country_code
    loc.state = data.region_name
    loc.city = data.city
    loc.business = ""
    loc.website = ""
    loc.zipcode = data.zip
    return loc



# Generated by Django 4.1.4 on 2024-01-11 18:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0018_userpasskey'),
        ('incident', '0013_rulecheck_is_required'),
    ]

    operations = [
        migrations.AddField(
            model_name='event',
            name='group',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='account.group'),
        ),
        migrations.AlterField(
            model_name='rulecheck',
            name='index',
            field=models.IntegerField(db_index=True, default=0),
        ),
    ]



from rest import settings
from rest import log
from objict import objict
import os
import subprocess

GIT_PROJECTS = settings.get("GIT_PROJECTS", None)
logger = log.getLogger("pushit", filename="pushit.log")


def updateCode(update_cmd, branch, update_user="ec2-user"):
    if settings.PUSHIT_TEST_SUDO:
        cmd = ["sudo", "-lU", update_user, update_cmd, branch]
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = process.communicate()
        if err.strip():
            logger.warning("WARNING: cannot run {}, we don't have sudo rights".format(update_cmd))
            logger.warning("WARNING: add {} to your sudo user".format(update_cmd))
            return
    # - lock it so we kill any updates in progress and start a new one
    cmd = ["sudo", "-u", update_user, update_cmd, branch]
    logger.info("updating...", cmd)
    subprocess.Popen(cmd, close_fds=True)


def parseGitLab(request):
    info = objict(vendor="gitlab")
    info.name = request.DATA.get("project.name")
    info.kind = request.DATA.get("object_kind")
    if "ref" in request.DATA:
        info.branch = request.DATA.get("ref").split('/')[-1]
    if info.kind == "merge_request":
        info.state = request.DATA.get("object_attributes.state", None)
        if info.state == "merged":
            info.kind = "merged"
    return info


def parseGithub(request):
    info = objict(vendor="github")
    info.name = request.DATA.get("repository.name")
    info.kind = request.DATA.getHeader("HTTP_X_GITHUB_EVENT")
    if "ref" in request.DATA:
        info.branch = request.DATA.get("ref").split('/')[-1]
    logger.info("github request", info)
    return info


def getProjectInfo(name, branch):
    return getProjectForBranch(GIT_PROJECTS.get(name, None), branch)


def getProjectForBranch(proj_info, branch):
    if proj_info is None:
        return None
    if isinstance(proj_info, list):
        for pi in proj_info:
            if pi["branch"] == branch:
                return pi
        return None
    if proj_info["branch"] == branch:
        return proj_info
    return None



from django.db import models as dm
from rest import models as rm
from rest import settings
from rest import helpers as rh
from medialib import models as medialib
import re
import mistune
from wiki.renderers import WikiRenderer
from wiki.renderers.mistune import task_list

WIKI_PAGE_VIEW_PERMS = settings.get("WIKI_PAGE_VIEW_PERMS", ["view_wiki", "edit_wiki"])
WIKI_PAGE_EDIT_PERMS = settings.get("WIKI_PAGE_EDIT_PERMS", ["edit_wiki"])


class Page(dm.Model, rm.RestModel, rm.MetaDataModel):
    """
    Blog (a collection of articles)
    """
    class RestMeta:
        SEARCH_FIELDS = ["title", "body"]
        SEARCH_TERMS = ["title", "body"]
        QUERY_FIELDS = ["all_fields", "parent__path"]
        VIEW_PERMS = WIKI_PAGE_VIEW_PERMS
        EDIT_PERMS = WIKI_PAGE_EDIT_PERMS
        UNIQUE_LOOKUP = ["path"]
        DEFAULT_SORT = "-order"
        CAN_DELETE = True
        GRAPHS = {
            "basic": {
                "fields": [
                    "id",
                    "path",
                    "created",
                    "modified",
                    "title",
                    "parent",
                    "order",
                    "path",
                    "slug",
                    "perms"
                ],
            },
            "default": {
                "fields": [
                    "id",
                    "path",
                    "created",
                    "modified",
                    "title",
                    "parent",
                    "order",
                    "slug",
                    "children_paths",
                    "body",
                    ("toHTML", "html")
                ],
                "recurse_into": ["media"],
                "graphs": {
                    "media": "default",
                    "member": "basic",
                    "parent": "basic"
                }
            },
            "rendered": {
                "graphs": {
                    "self": "default",
                    "member": "basic"
                }
            },
            "list": {
                "graphs": {
                    "self": "basic",
                    "member": "basic"
                }
            },
            "toc_child": {
                "graphs": {
                    "self": "basic",
                    "parent": "basic"
                }
            },
            "toc": {
                "extra": ["children", "metadata"],
                "graphs": {
                    "self": "basic",
                    "children": "toc_child"
                }
            },
        }

    created = dm.DateTimeField(auto_now_add=True, editable=False)
    modified = dm.DateTimeField(db_index=True, auto_now=True)

    member = dm.ForeignKey(
        "account.Member", related_name="+",
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    group = dm.ForeignKey(
        "account.Group", related_name="+", 
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    parent = dm.ForeignKey(
        "Page", related_name="children",
        default=None, null=True, blank=True,
        on_delete=dm.CASCADE)

    order = dm.IntegerField(default=0, blank=True)

    title = dm.CharField(max_length=255)
    path = dm.CharField(max_length=255, db_index=True)
    slug = dm.SlugField(db_index=True)
    perms = dm.CharField(max_length=255, db_index=True, null=True, default=None)

    body = dm.TextField(blank=True)

    is_active = dm.BooleanField(blank=True, default=True)

    @property
    def children_paths(self):
        return list(self.children.all().values_list("path", flat=True))

    def upload__file(self, fobj, name):
        if fobj is None:
            return

        request = self.getActiveRequest()
        if not self.group:
            lib = request.member.getMediaLibrary("wiki")
        else:
            lib = self.group.getMediaLibrary("wiki")
        kind = medialib.MediaItem.guessMediaKind(fobj)
        media = medialib.MediaItem(library=lib, name=fobj.name, member=request.member, kind=kind, newfile=fobj)
        media.save()
        wmedia = PageMedia(media=media, entry=self, group=self.group, member=request.member)
        wmedia.save()

    def set_remove_media(self, value):
        wmedia = PageMedia.objects.filter(pk=value).last()
        if wmedia:
            wmedia.delete()
    
    def set_slug(self, value):
        sanitized_slug = value.strip().lower().replace(' ', '_')
        sanitized_slug = re.sub(r'\W+', '', sanitized_slug)
        sanitized_slug = re.sub(r'__+', '_', sanitized_slug).strip('_')
        if self.slug != sanitized_slug:
            self._recordRestChange("slug", self.slug)
            self.slug = sanitized_slug
            # confirm slug is unique
            qset = Page.objects.filter(parent=self.parent, slug=sanitized_slug)
            if qset.count():
                self.slug = f"{sanitized_slug}_{qset.count()+1}"

    def on_rest_can_get(self, request):
        if request is None:
            return True
        if self.perms:
            perms = [p.strip() for p in self.perms.split(',')]
        elif self.parent and self.parent.perms:
            perms = [p.strip() for p in self.parent.perms.split(',')]
        else:
            perms = getattr(self.RestMeta, "VIEW_PERMS", None)
        if perms:
            if "public" in perms:
                return True
            if "owner" in perms and self.checkIsOwner(request.member):
                return True
            # we need to check if this user has permission
            group_field = "group"
            status, error, code = rh.requestHasPerms(request, perms, getattr(self, group_field, None))
            if not status:
                return False
        return True

    @classmethod
    def on_rest_list_ready(cls, request, qset=None):
        # need to first sort our list
        qset = cls.restListSort(request, qset)
        out = []
        for page in qset:
            if page.on_rest_can_get(request):
                out.append(page)
        return out

    def on_rest_pre_save(self, request):
        if not self.slug:
            self.set_slug(self.title)
        self.generatePath()

    def on_rest_saved(self, request, is_new=False):
        if self.hasFieldChanged("slug"):
            self.generatePath(True)

    def generatePath(self, propagate=False):
        paths = [self.slug]
        parent = self.parent
        while parent is not None:
            paths.append(parent.slug)
            parent = parent.parent
        paths.reverse()
        path = "/".join(paths)
        if self.path != path:
            self._recordRestChange("path", self.path)
            self.path = path
            if propagate:
                self.save()
        if propagate:
            for child in self.children.all():
                child.generatePath(propagate=True)
        return self.path

    def toHTML(self):
        renderer = WikiRenderer()
        md_engine = mistune.create_markdown(
            renderer=renderer,
            plugins=['def_list', 'url', 'abbr'])
        md_engine.use(task_list.plugin_task_lists)
        md = md_engine(self.body)
        return {
            "toc": renderer.render_toc(),
            "body": md
        }

    @classmethod
    def path_first_existing_parent(cls, path):
        paths = path.split("/")
        while len(paths):
            paths.pop()
            parent = cls.objects.filter(path="/".join(paths)).first()
            if parent is not None:
                return parent
        return None

    @classmethod
    def path_parent(cls, path):
        paths = path.split("/")
        paths.pop()
        return cls.objects.filter(path="/".join(paths)).first()


class PageMetaData(rm.MetaDataBase):
    parent = dm.ForeignKey(Page, related_name="properties", on_delete=dm.CASCADE)


class PageMedia(dm.Model, rm.RestModel):
    class RestMeta:
        VIEW_PERMS = WIKI_PAGE_VIEW_PERMS
        EDIT_PERMS = WIKI_PAGE_EDIT_PERMS
        GRAPHS = {
            "basic": {
                "graphs": {
                    "media": "default"
                }
            },
            "default": {
                "graphs": {
                    "media": "default"
                }
            }
        }
    created = dm.DateTimeField(auto_now_add=True, editable=False)

    member = dm.ForeignKey(
        "account.Member", related_name="+",
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    group = dm.ForeignKey(
        "account.Group", related_name="+", 
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    entry = dm.ForeignKey(Page, related_name="media_library", on_delete=dm.CASCADE)
    media = dm.ForeignKey("medialib.MediaItem", related_name="+", on_delete=dm.CASCADE)



# Generated by Django 4.1.4 on 2023-12-19 03:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incident', '0011_ticket'),
    ]

    operations = [
        migrations.AddField(
            model_name='rule',
            name='match_by',
            field=models.IntegerField(default=0),
        ),
    ]




from django.conf import settings

import os
import errno
import urllib.request, urllib.parse, urllib.error
import shutil
from urllib.parse import urlparse


def toString(value):
    if isinstance(value, bytes):
        value = value.decode()
    elif isinstance(value, bytearray):
        value = value.decode("utf-8")
    elif isinstance(value, (int, float)):
        value = str(value)
    return value


def upload(url, fp, background=False):
    u = urlparse(url)
    fn = os.path.join(settings.MEDIA_ROOT, u.path.lstrip("/"))
    dir = os.path.dirname(fn)
    try:
        os.makedirs(dir)
    except OSError as exc:
        if exc.errno == errno.EEXIST:
            pass
        else:
            raise
    ofp = open(fn, "wb")
    buf = fp.read(10240)
    while buf:
        ofp.write(buf)
        buf = fp.read(10240)
    ofp.close()

def view_url(url, expires=600, is_secure=False):
    u = urlparse(url)
    return settings.MEDIA_URL + u.path.lstrip("/")

def exists(url):
    u = urlparse(url)
    fn = os.path.join(settings.MEDIA_ROOT, u.path.lstrip("/"))
    return os.access(fn, os.R_OK)

def get_file(url, fp):
    u = urlparse(url)
    fn = os.path.join(settings.MEDIA_ROOT, u.path.lstrip("/"))
    ifp = open(fn, "rb")
    buf = ifp.read(10240)
    while buf:
        fp.write(buf)
        buf = ifp.read(10240)
    ifp.close()

def delete(url):
    u = urlparse(url)
    fn = os.path.join(settings.MEDIA_ROOT, u.path.lstrip("/"))
    try:
        if fn[-1] == "/":
            shutil.rmtree(fn[:-1])
        else:
            os.unlink(fn)
    except OSError:
        pass



# Generated by Django 4.0.2 on 2022-02-22 01:25

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import rest.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='GeoIP',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('hostname', models.CharField(blank=True, default=None, max_length=255, null=True)),
                ('ip', models.CharField(max_length=64)),
                ('isp', models.CharField(max_length=84)),
                ('city', models.CharField(blank=True, max_length=64, null=True)),
                ('state', models.CharField(blank=True, max_length=64, null=True)),
                ('country', models.CharField(blank=True, max_length=64, null=True)),
                ('postal', models.CharField(blank=True, max_length=32, null=True)),
                ('lat', models.FloatField(blank=True, default=0.0)),
                ('lng', models.FloatField(blank=True, default=0.0)),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='GeoLocation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('city', models.CharField(blank=True, max_length=64, null=True)),
                ('county', models.CharField(max_length=64)),
                ('state', models.CharField(max_length=2)),
                ('zip', models.CharField(max_length=32)),
                ('country', models.CharField(max_length=64)),
                ('area_codes', models.CharField(max_length=64)),
                ('timezone', models.CharField(max_length=32)),
                ('kind', models.CharField(max_length=32)),
                ('lat', models.FloatField(blank=True, default=0.0)),
                ('lng', models.FloatField(blank=True, default=0.0)),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='GeoTrack',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uid', models.CharField(db_index=True, default=None, max_length=124, null=True)),
                ('name', models.CharField(max_length=124)),
                ('kind', models.CharField(max_length=64)),
                ('ref_id', models.IntegerField(blank=True, default=0)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('owner', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='GeoPosition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uid', models.CharField(db_index=True, default=None, max_length=124, null=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('lat', models.FloatField(default=0.0)),
                ('lng', models.FloatField(default=0.0)),
                ('elevation', models.FloatField(blank=True, default=0.0)),
                ('accuracy', models.FloatField(blank=True, default=0.0)),
                ('altitude', models.FloatField(blank=True, default=0.0)),
                ('altitudeAccuracy', models.FloatField(blank=True, default=0.0)),
                ('heading', models.FloatField(blank=True, default=0.0)),
                ('speed', models.FloatField(blank=True, default=0.0)),
                ('track', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='positions', to='location.geotrack')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='GeoIPLocation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('obj_id', models.IntegerField(blank=True, default=0)),
                ('obj_kind', models.CharField(blank=True, default='', max_length=64, null=True)),
                ('ip', models.CharField(max_length=32)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('city', models.CharField(blank=True, max_length=64, null=True)),
                ('region', models.CharField(max_length=64)),
                ('country', models.CharField(max_length=64)),
                ('area_codes', models.CharField(max_length=64)),
                ('timezone', models.CharField(max_length=32)),
                ('kind', models.CharField(max_length=32)),
                ('lat', models.FloatField(blank=True, default=0.0)),
                ('lng', models.FloatField(blank=True, default=0.0)),
                ('owner', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='locations', to=settings.AUTH_USER_MODEL)),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='Address',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('modified', models.DateTimeField(auto_now=True)),
                ('line1', models.CharField(blank=True, default=None, max_length=255, null=True)),
                ('line2', models.CharField(blank=True, default=None, max_length=255, null=True)),
                ('city', models.CharField(blank=True, default=None, max_length=127, null=True)),
                ('state', models.CharField(blank=True, default=None, max_length=127, null=True)),
                ('county', models.CharField(blank=True, default=None, max_length=127, null=True)),
                ('country', models.CharField(blank=True, default=None, max_length=16, null=True)),
                ('postalcode', models.CharField(blank=True, default=None, max_length=32, null=True)),
                ('lat', models.FloatField(blank=True, default=0.0)),
                ('lng', models.FloatField(blank=True, default=0.0)),
                ('modified_by', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
    ]



# Generated by Django 4.1.3 on 2022-12-16 17:12

from django.db import migrations, models
import django.db.models.deletion
import rest.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('medialib', '0001_initial'),
        ('location', '0002_geoip_subnet_alter_geoip_ip'),
        ('account', '0012_settings_settingsmetadata'),
    ]

    operations = [
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('reporter_ip', models.CharField(blank=True, db_index=True, default=None, max_length=16, null=True)),
                ('hostname', models.CharField(blank=True, db_index=True, default=None, max_length=255, null=True)),
                ('description', models.CharField(max_length=84)),
                ('details', models.TextField(default=None, null=True)),
                ('level', models.IntegerField(db_index=True, default=0)),
                ('category', models.CharField(db_index=True, max_length=124)),
            ],
            bases=(models.Model, rest.models.RestModel, rest.models.MetaDataModel),
        ),
        migrations.CreateModel(
            name='Incident',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('description', models.CharField(max_length=200)),
                ('priority', models.IntegerField(default=0)),
                ('state', models.IntegerField(default=0)),
                ('assigned_to', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to='account.member')),
                ('group', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to='account.group')),
            ],
            bases=(models.Model, rest.models.RestModel, rest.models.MetaDataModel),
        ),
        migrations.CreateModel(
            name='Rule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('category', models.CharField(db_index=True, max_length=200)),
                ('priority', models.IntegerField(db_index=True, default=10)),
                ('bundle', models.IntegerField(default=0)),
                ('created_by', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to='account.member')),
                ('group', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to='account.group')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='ServerOssecAlert',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('when', models.DateTimeField(db_index=True)),
                ('alert_id', models.CharField(default=None, max_length=64, null=True)),
                ('rule_id', models.CharField(db_index=True, default=None, max_length=64, null=True)),
                ('text', models.TextField(default=None, null=True)),
                ('action', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('ext_ip', models.CharField(blank=True, default=None, max_length=32, null=True)),
                ('src_ip', models.CharField(blank=True, default=None, max_length=32, null=True)),
                ('hostname', models.CharField(blank=True, db_index=True, default=None, max_length=128, null=True)),
                ('ip', models.CharField(blank=True, default=None, max_length=32, null=True)),
                ('username', models.CharField(default=None, max_length=64, null=True)),
                ('username2', models.CharField(default=None, max_length=64, null=True)),
                ('level', models.IntegerField(default=0)),
                ('title', models.CharField(blank=True, default=None, max_length=200, null=True)),
                ('geoip', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='location.geoip')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='RuleCheck',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('index', models.IntegerField(default=0)),
                ('comparator', models.CharField(default='==', max_length=32)),
                ('field_name', models.CharField(default=None, max_length=124, null=True)),
                ('value', models.CharField(default='', max_length=124)),
                ('value_type', models.IntegerField(default=0)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='incident.rule')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='IncidentMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='incident.incident')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='IncidentHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('kind', models.CharField(blank=True, db_index=True, default=None, max_length=80, null=True)),
                ('state', models.IntegerField(default=0)),
                ('priority', models.IntegerField(default=0)),
                ('note', models.TextField(blank=True, default=None, null=True)),
                ('by', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='account.member')),
                ('group', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='account.group')),
                ('media', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='medialib.mediaitem')),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='history', to='incident.incident')),
                ('to', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='account.member')),
            ],
            options={
                'ordering': ['-created'],
            },
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.AddField(
            model_name='incident',
            name='rule',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to='incident.rule'),
        ),
        migrations.CreateModel(
            name='EventMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='incident.event')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='event',
            name='incident',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='events', to='incident.incident'),
        ),
    ]
